# app.py
from flask import Flask, render_template, request, jsonify
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import func
from datetime import datetime
import pdfkit
import json

app = Flask(__name__)

# Updated SQLAlchemy connection string
app.config['SQLALCHEMY_DATABASE_URI'] = 'mysql+pymysql://root:@localhost/stops_db'
app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
db = SQLAlchemy(app)
@app.route('/')
def index():
    return render_template('index.html')

class Stop(db.Model):
    __tablename__ = 'stops'
    id = db.Column(db.Integer, primary_key=True)
    sloid = db.Column(db.String(100))
    uic_ref = db.Column(db.String(100), index=True)  # Merged field from ATLAS number and OSM uic_ref
    atlas_designation = db.Column(db.String(255))
    atlas_designation_official = db.Column(db.String(255))
    atlas_business_org_abbr = db.Column(db.String(100))
    osm_node_id = db.Column(db.String(100))
    osm_local_ref = db.Column(db.String(100))
    osm_network = db.Column(db.String(255))
    osm_operator = db.Column(db.String(255))
    osm_public_transport = db.Column(db.String(255))
    stop_type = db.Column(db.String(50))  # 'matched', 'unmatched', 'osm'
    match_type = db.Column(db.String(50))
    atlas_lat = db.Column(db.Float)
    atlas_lon = db.Column(db.Float)
    osm_name = db.Column(db.String(255))
    osm_uic_name = db.Column(db.String(255))
    osm_lat = db.Column(db.Float)
    osm_lon = db.Column(db.Float)
    distance_m = db.Column(db.Float)
    routes_atlas = db.Column(db.JSON)  # List of {route_id, direction_id, route_short_name} for ATLAS
    routes_osm = db.Column(db.JSON)  # List of {route_id, direction_id, route_short_name} for OSM

# Create tables within application context when the app runs
with app.app_context():
    db.create_all()

# ----------------------------
# API Endpoint: /api/data
# ----------------------------
@app.route('/api/data', methods=['GET'])
def get_data():
    try:
        # Parse viewport parameters
        try:
            min_lat = float(request.args.get('min_lat'))
            max_lat = float(request.args.get('max_lat'))
            min_lon = float(request.args.get('min_lon'))
            max_lon = float(request.args.get('max_lon'))
        except Exception as e:
            return jsonify({"error": "Missing or invalid viewport parameters."}), 400

        # Parse filter and pagination parameters
        stop_filter = request.args.get('stop_filter', None)
        match_method = request.args.get('match_method', 'all')
        topN = request.args.get('topN', None)
        
        # New: Get station filter values (comma-separated list)
        station_filter = request.args.get('station_filter', None)
        filter_type = request.args.get('filter_type', 'station')
        
        try:
            offset = int(request.args.get('offset', 0))
            limit = int(request.args.get('limit', 500))
        except:
            offset = 0
            limit = 500

        # Build base query with viewport bounds
        query = Stop.query.filter(
            func.coalesce(Stop.atlas_lat, Stop.osm_lat) >= min_lat,
            func.coalesce(Stop.atlas_lat, Stop.osm_lat) <= max_lat,
            func.coalesce(Stop.atlas_lon, Stop.osm_lon) >= min_lon,
            func.coalesce(Stop.atlas_lon, Stop.osm_lon) <= max_lon
        )

        # Apply stop type filter
        if stop_filter and stop_filter.lower() != 'all':
            types = [t.strip() for t in stop_filter.split(',') if t.strip()]
            if types:
                query = query.filter(Stop.stop_type.in_(types))

        # Apply match method filter
        if match_method != 'all':
            query = apply_match_method_filter(query, match_method)
            
        # Apply station/node filter if provided
        if station_filter:
            filter_values = [val.strip() for val in station_filter.split(',') if val.strip()]
            filter_types = request.args.get('filter_types', '').split(',')
            route_directions = request.args.get('route_directions', '').split(',')
            
            # Make sure we have types for all values (default to 'station' if missing)
            while len(filter_types) < len(filter_values):
                filter_types.append('station')
            
            # Make sure we have directions for all values (default to empty string if missing)
            while len(route_directions) < len(filter_values):
                route_directions.append('')
            
            if filter_values:
                # Build combined OR conditions for all filter values
                combined_conditions = []
                
                for i, value in enumerate(filter_values):
                    filter_type = filter_types[i].strip() if i < len(filter_types) else 'station'
                    direction = route_directions[i].strip() if i < len(route_directions) else ''
                    
                    if filter_type == 'atlas':
                        # Filter by ATLAS SloidID
                        combined_conditions.append(Stop.sloid.like(f'%{value}%'))
                    elif filter_type == 'osm':
                        # Filter by OSM Node ID
                        combined_conditions.append(Stop.osm_node_id.like(f'%{value}%'))
                    elif filter_type == 'route':
                        # For route filtering, get all stops that belong to this route
                        route_stops = get_stops_for_route(value, direction if direction else None)
                        
                        route_conditions = []
                        
                        # Add atlas_sloids to conditions
                        if route_stops['atlas_sloids']:
                            route_conditions.append(Stop.sloid.in_(route_stops['atlas_sloids']))
                        
                        # Add osm_nodes to conditions
                        if route_stops['osm_nodes']:
                            route_conditions.append(Stop.osm_node_id.in_(route_stops['osm_nodes']))
                        
                        # Combine with OR if we have both types
                        if route_conditions:
                            combined_conditions.append(db.or_(*route_conditions))
                    else:
                        # Filter by UIC Number (default)
                        combined_conditions.append(Stop.uic_ref.like(f'%{value}%'))
                
                if combined_conditions:
                    query = query.filter(db.or_(*combined_conditions))

        # Apply topN filter or pagination
        if stop_filter == 'matched' and topN:
            try:
                topN_int = int(topN)
                query = query.order_by(Stop.distance_m.desc()).limit(topN_int)
            except Exception as e:
                query = query.offset(offset).limit(limit)
        else:
            query = query.offset(offset).limit(limit)

        # Execute query
        stops = query.all()
        
        # Process results
        atlas_to_osm_matches = {}
        regular_stops = []
        
        for stop in stops:
            if stop.stop_type == 'matched' and stop.sloid:
                # Group ATLAS stops with multiple OSM matches
                if stop.sloid not in atlas_to_osm_matches:
                    atlas_to_osm_matches[stop.sloid] = {
                        "id": stop.id,
                        "sloid": stop.sloid,
                        "stop_type": stop.stop_type,
                        "match_type": stop.match_type,
                        "uic_ref": stop.uic_ref,
                        "atlas_designation": stop.atlas_designation,
                        "atlas_designation_official": stop.atlas_designation_official,
                        "atlas_business_org_abbr": stop.atlas_business_org_abbr,
                        "atlas_lat": stop.atlas_lat,
                        "atlas_lon": stop.atlas_lon,
                        "lat": stop.atlas_lat if stop.atlas_lat is not None else stop.osm_lat,
                        "lon": stop.atlas_lon if stop.atlas_lon is not None else stop.osm_lon,
                        "routes_atlas": stop.routes_atlas,
                        "osm_matches": []
                    }
                
                # Add OSM match data
                atlas_to_osm_matches[stop.sloid]["osm_matches"].append({
                    "osm_id": stop.id,
                    "osm_node_id": stop.osm_node_id,
                    "osm_local_ref": stop.osm_local_ref,
                    "osm_network": stop.osm_network,
                    "osm_operator": stop.osm_operator,
                    "osm_public_transport": stop.osm_public_transport,
                    "osm_name": stop.osm_name,
                    "osm_uic_name": stop.osm_uic_name,
                    "osm_lat": stop.osm_lat,
                    "osm_lon": stop.osm_lon,
                    "distance_m": stop.distance_m,
                    "routes_osm": stop.routes_osm,
                    "match_type": stop.match_type
                })
            else:
                # Handle regular stops
                lat = stop.atlas_lat if stop.atlas_lat is not None else stop.osm_lat
                lon = stop.atlas_lon if stop.atlas_lon is not None else stop.osm_lon
                regular_stops.append({
                    "id": stop.id,
                    "sloid": stop.sloid,
                    "stop_type": stop.stop_type,
                    "match_type": stop.match_type,
                    "uic_ref": stop.uic_ref,
                    "atlas_designation": stop.atlas_designation,
                    "atlas_designation_official": stop.atlas_designation_official,
                    "atlas_business_org_abbr": stop.atlas_business_org_abbr,
                    "osm_node_id": stop.osm_node_id,
                    "osm_local_ref": stop.osm_local_ref,
                    "osm_network": stop.osm_network,
                    "osm_operator": stop.osm_operator,
                    "osm_public_transport": stop.osm_public_transport,
                    "atlas_lat": stop.atlas_lat if stop.atlas_lat is not None else stop.osm_lat,
                    "atlas_lon": stop.atlas_lon if stop.atlas_lon is not None else stop.osm_lon,
                    "osm_name": stop.osm_name,
                    "osm_uic_name": stop.osm_uic_name,
                    "osm_lat": stop.osm_lat,
                    "osm_lon": stop.osm_lon,
                    "distance_m": stop.distance_m,
                    "lat": lat,
                    "lon": lon,
                    "routes_atlas": stop.routes_atlas,
                    "routes_osm": stop.routes_osm
                })
        
        # Combine and return results
        combined_stops = regular_stops + list(atlas_to_osm_matches.values())
        return jsonify(combined_stops)
    except Exception as e:
        return jsonify({"error": str(e)}), 500

# ----------------------------
# API Endpoint: /api/save
# ----------------------------
@app.route('/api/save', methods=['POST'])
def save_changes():
    try:
        changes = request.json
        manual_matches = changes.get("manualMatches", [])
        for pair in manual_matches:
            atlas_id = pair.get("atlas_id")
            osm_id = pair.get("osm_id")
            if atlas_id and osm_id:
                atlas_stop = Stop.query.get(atlas_id)
                osm_stop = Stop.query.get(osm_id)
                if atlas_stop and osm_stop:
                    atlas_stop.stop_type = 'matched'
                    osm_stop.stop_type = 'matched'
                    atlas_stop.match_type = 'manual'
                    osm_stop.match_type = 'manual'
                    db.session.add(atlas_stop)
                    db.session.add(osm_stop)
        db.session.commit()
        with open("changes.json", "w", encoding="utf-8") as f:
            json.dump(changes, f, indent=2, ensure_ascii=False)
        return jsonify({"status": "success", "message": "Changes saved."})
    except Exception as e:
        return jsonify({"status": "error", "message": str(e)}), 500

# ----------------------------
# API Endpoint: /api/generate_report
# ----------------------------
@app.route('/api/generate_report', methods=['GET'])
def generate_report():
    try:
        limit = int(request.args.get('limit', 10))
        sort = request.args.get('sort', 'distance_desc')
        report_type = request.args.get('report_type', 'top_matches')
        
        query = Stop.query.filter(Stop.stop_type == 'matched')
        
        # Apply specific filtering based on report type
        if report_type == 'exact_matches':
            query = query.filter(Stop.match_type == 'exact')
        elif report_type == 'name_matches':
            query = query.filter(Stop.match_type == 'name')
        # top_matches includes all match types
        
        # Apply sorting based on sort parameter
        if sort == 'id_asc':
            query = query.order_by(Stop.id.asc())
        elif sort == 'distance_asc':
            query = query.filter(Stop.distance_m != None).order_by(Stop.distance_m.asc())
        elif sort == 'distance_desc':
            query = query.filter(Stop.distance_m != None).order_by(Stop.distance_m.desc())
        else:
            query = query.order_by(Stop.id.asc())
        
        top_matches = query.limit(limit).all()
        
        # Create descriptive report title based on report type
        report_title = "OSM & ATLAS Matching Report"
        if report_type == 'exact_matches':
            report_title = "Top Exact Matches by Distance"
        elif report_type == 'name_matches':
            report_title = "Top Name Matches by Distance"
        elif report_type == 'top_matches':
            report_title = "Top Matches by Distance"
        
        report_html = render_template('report.html', 
                                     top_matches=top_matches, 
                                     generated_at=datetime.now(), 
                                     sort_order=sort,
                                     report_title=report_title)
        pdf = pdfkit.from_string(report_html, False)
        response = app.response_class(pdf, mimetype='application/pdf')
        response.headers["Content-Disposition"] = "attachment; filename=report.pdf"
        return response
    except Exception as e:
        return jsonify({"status": "error", "message": str(e)}), 500

# ----------------------------
# API Endpoint: /api/search
# ----------------------------
@app.route('/api/search', methods=['GET'])
def search():
    query_str = request.args.get('q', '').lower()
    results = {"osm": [], "atlas": []}
    if query_str:
        matched_stops = Stop.query.filter(Stop.stop_type == 'matched').all()
        for stop in matched_stops:
            if (query_str in (stop.atlas_designation or "").lower() or
                query_str in (stop.atlas_designation_official or "").lower() or
                query_str in (stop.uic_ref or "").lower() or
                query_str in (stop.atlas_business_org_abbr or "").lower() or
                query_str in (stop.osm_name or "").lower() or
                query_str in (stop.osm_local_ref or "").lower() or
                query_str in (stop.osm_network or "").lower() or
                query_str in (stop.osm_operator or "").lower() or
                query_str in (stop.osm_uic_name or "").lower()):
                results['atlas'].append({
                    "sloid": stop.sloid,
                    "stop_type": stop.stop_type,
                    "atlas_lat": stop.atlas_lat,
                    "atlas_lon": stop.atlas_lon,
                    "atlas_business_org_abbr": stop.atlas_business_org_abbr,
                    "osm_lat": stop.osm_lat,
                    "osm_lon": stop.osm_lon,
                    "osm_network": stop.osm_network,
                    "osm_operator": stop.osm_operator,
                    "match_type": stop.match_type,
                    "atlas_designation": stop.atlas_designation,
                    "atlas_designation_official": stop.atlas_designation_official,
                    "uic_ref": stop.uic_ref,
                    "osm_node_id": stop.osm_node_id,
                    "osm_local_ref": stop.osm_local_ref,
                    "osm_uic_name": stop.osm_uic_name
                })
        unmatched_stops = Stop.query.filter(Stop.stop_type == 'unmatched').all()
        for stop in unmatched_stops:
            if (query_str in (stop.atlas_designation or "").lower() or
                query_str in (stop.atlas_designation_official or "").lower() or
                query_str in (stop.atlas_business_org_abbr or "").lower() or
                query_str in (stop.uic_ref or "").lower()):
                results['atlas'].append({
                    "sloid": stop.sloid,
                    "stop_type": stop.stop_type,
                    "atlas_lat": stop.atlas_lat,
                    "atlas_lon": stop.atlas_lon,
                    "atlas_business_org_abbr": stop.atlas_business_org_abbr,
                    "match_type": stop.match_type,
                    "atlas_designation": stop.atlas_designation,
                    "atlas_designation_official": stop.atlas_designation_official,
                    "uic_ref": stop.uic_ref
                })
    return jsonify(results)

# ----------------------------
# API Endpoint: /api/top_matches
# ----------------------------
@app.route('/api/top_matches', methods=['GET'])
def get_top_matches():
    try:
        # Parse parameters
        limit = int(request.args.get('limit', 10))
        match_method = request.args.get('match_method', None)
        
        # Get station filter values (comma-separated list)
        station_filter = request.args.get('station_filter', None)
        filter_types = request.args.get('filter_types', '').split(',')
        route_directions = request.args.get('route_directions', '').split(',')
        
        # Build base query
        query = Stop.query.filter(Stop.stop_type=='matched')
        
        # Apply match method filtering if provided
        if match_method:
            query = apply_match_method_filter(query, match_method)
            
        # Apply station/node filter if provided
        if station_filter:
            filter_values = [val.strip() for val in station_filter.split(',') if val.strip()]
            
            # Make sure we have types for all values (default to 'station' if missing)
            while len(filter_types) < len(filter_values):
                filter_types.append('station')
            
            # Make sure we have directions for all values (default to empty string if missing)
            while len(route_directions) < len(filter_values):
                route_directions.append('')
            
            if filter_values:
                # Build combined OR conditions for all filter values
                combined_conditions = []
                
                for i, value in enumerate(filter_values):
                    filter_type = filter_types[i].strip() if i < len(filter_types) else 'station'
                    direction = route_directions[i].strip() if i < len(route_directions) else ''
                    
                    if filter_type == 'atlas':
                        # Filter by ATLAS SloidID
                        combined_conditions.append(Stop.sloid.like(f'%{value}%'))
                    elif filter_type == 'osm':
                        # Filter by OSM Node ID
                        combined_conditions.append(Stop.osm_node_id.like(f'%{value}%'))
                    elif filter_type == 'route':
                        # For route filtering, get all stops that belong to this route
                        route_stops = get_stops_for_route(value, direction if direction else None)
                        
                        route_conditions = []
                        
                        # Add atlas_sloids to conditions
                        if route_stops['atlas_sloids']:
                            route_conditions.append(Stop.sloid.in_(route_stops['atlas_sloids']))
                        
                        # Add osm_nodes to conditions
                        if route_stops['osm_nodes']:
                            route_conditions.append(Stop.osm_node_id.in_(route_stops['osm_nodes']))
                        
                        # Combine with OR if we have both types
                        if route_conditions:
                            combined_conditions.append(db.or_(*route_conditions))
                    else:
                        # Filter by UIC Number (default)
                        combined_conditions.append(Stop.uic_ref.like(f'%{value}%'))
                
                if combined_conditions:
                    query = query.filter(db.or_(*combined_conditions))
        
        # Get results sorted by distance
        stops = query.filter(Stop.distance_m != None) \
                    .order_by(Stop.distance_m.desc()) \
                    .limit(limit) \
                    .all()
        
        # Format response data
        stops_data = [format_stop_data(stop) for stop in stops]
        return jsonify(stops_data)
    except Exception as e:
        return jsonify({"error": str(e)}), 500

# Helper function to apply match method filtering
def apply_match_method_filter(query, match_method):
    methods = [m.strip() for m in match_method.split(',') if m.strip()]
    if not methods:
        return query
        
    # Handle special cases for route matching
    like_conditions = []
    exact_methods = []
    
    for method in methods:
        if method.endswith('_'):
            # Handle wildcard matching for route_matching_1_, route_matching_2_, etc.
            like_conditions.append(Stop.match_type.like(f'{method}%'))
        else:
            exact_methods.append(method)
    
    # Combine conditions with OR logic
    filter_conditions = like_conditions.copy()
    if exact_methods:
        filter_conditions.append(Stop.match_type.in_(exact_methods))
    
    if filter_conditions:
        return query.filter(db.or_(*filter_conditions))
    return query

# Helper function to format stop data for API response
def format_stop_data(stop):
    return {
        "id": stop.id,
        "sloid": stop.sloid,
        "stop_type": stop.stop_type,
        "match_type": stop.match_type,
        "atlas_lat": stop.atlas_lat if stop.atlas_lat is not None else stop.osm_lat,
        "atlas_lon": stop.atlas_lon if stop.atlas_lon is not None else stop.osm_lon,
        "atlas_business_org_abbr": stop.atlas_business_org_abbr,
        "osm_lat": stop.osm_lat,
        "osm_lon": stop.osm_lon,
        "osm_network": stop.osm_network,
        "osm_operator": stop.osm_operator,
        "osm_public_transport": stop.osm_public_transport,
        "distance_m": stop.distance_m,
        "atlas_designation": stop.atlas_designation,
        "atlas_designation_official": stop.atlas_designation_official,
        "uic_ref": stop.uic_ref,
        "osm_node_id": stop.osm_node_id,
        "osm_local_ref": stop.osm_local_ref,
        "osm_name": stop.osm_name,
        "osm_uic_name": stop.osm_uic_name,
        "routes_atlas": stop.routes_atlas,
        "routes_osm": stop.routes_osm
    }

# New route query functionality
def get_stops_for_route(route_id, direction=None):
    """
    Gets all stops (both OSM and ATLAS) that belong to a route.
    
    Args:
        route_id: The route ID to filter by
        direction: Optional direction ID to filter by
    
    Returns:
        A list of stop IDs (both ATLAS and OSM)
    """
    try:
        # Build the SQL query based on whether we're filtering by direction
        sql_query = """
            SELECT 
                osm_nodes_json, 
                atlas_sloids_json 
            FROM 
                routes_and_directions 
            WHERE 
                (osm_route_id LIKE :route_id 
                OR atlas_route_id LIKE :route_id)
        """
        
        params = {"route_id": f'%{route_id}%'}
        
        # Add direction filter if provided
        if direction:
            sql_query += " AND direction_id = :direction"
            params["direction"] = direction
        
        app.logger.info(f"Executing query for route {route_id} with direction {direction if direction else 'None'}")
        
        # Execute the query
        route_entries = db.session.execute(db.text(sql_query), params).fetchall()
        
        # Extract all OSM nodes and ATLAS sloids
        osm_nodes = []
        atlas_sloids = []
        
        for entry in route_entries:
            # The tuple contains the columns in order: osm_nodes_json, atlas_sloids_json
            osm_nodes_json = entry[0]  # First column
            atlas_sloids_json = entry[1]  # Second column
            
            # Extract OSM nodes
            if osm_nodes_json:
                try:
                    # Handle string or JSON object
                    if isinstance(osm_nodes_json, str):
                        osm_nodes_list = json.loads(osm_nodes_json)
                    else:
                        osm_nodes_list = osm_nodes_json
                    
                    if isinstance(osm_nodes_list, list):
                        osm_nodes.extend(osm_nodes_list)
                    else:
                        app.logger.warning(f"Expected list of OSM nodes but got: {type(osm_nodes_list)}")
                except Exception as e:
                    app.logger.error(f"Error parsing OSM nodes JSON: {e}")
            
            # Extract ATLAS sloids
            if atlas_sloids_json:
                try:
                    # Handle string or JSON object
                    if isinstance(atlas_sloids_json, str):
                        atlas_sloids_list = json.loads(atlas_sloids_json)
                    else:
                        atlas_sloids_list = atlas_sloids_json
                    
                    if isinstance(atlas_sloids_list, list):
                        atlas_sloids.extend(atlas_sloids_list)
                    else:
                        app.logger.warning(f"Expected list of ATLAS sloids but got: {type(atlas_sloids_list)}")
                except Exception as e:
                    app.logger.error(f"Error parsing ATLAS sloids JSON: {e}")
        
        # Log what we found for debugging
        app.logger.info(f"Found {len(osm_nodes)} OSM nodes and {len(atlas_sloids)} ATLAS sloids for route {route_id}" + 
                      (f" with direction {direction}" if direction else ""))
        
        return {
            'osm_nodes': list(set(osm_nodes)),  # Remove duplicates
            'atlas_sloids': list(set(atlas_sloids))  # Remove duplicates
        }
    except Exception as e:
        app.logger.error(f"Error retrieving stops for route {route_id}: {e}")
        return {'osm_nodes': [], 'atlas_sloids': []}

# ---------------------------
# API Endpoint: /api/route_stops
# ---------------------------
@app.route('/api/route_stops', methods=['GET'])
def get_route_stops():
    """API endpoint to get all stops for a given route"""
    route_id = request.args.get('route_id')
    direction = request.args.get('direction')
    
    if not route_id:
        return jsonify({'error': 'No route ID provided'}), 400
    
    stops = get_stops_for_route(route_id, direction)
    return jsonify(stops)

if __name__ == '__main__':
    app.run(debug=True)
