"""
This script analyzes the ATLAS stops data to find duplicates based on 'number' and 'designation' columns.
Duplicates are saved to a CSV file and statistics are reported.
Only entries with non-empty designation values are considered for duplication.
"""
import pandas as pd

def find_duplicates():
    # Load the ATLAS stops data
    print("Loading ATLAS stops data...")
    atlas_data = pd.read_csv("stops_ATLAS.csv", sep=';')
    
    # Count initial number of entries
    total_entries = len(atlas_data)
    print(f"Loaded {total_entries} ATLAS stops.")
    
    # Filter out entries with empty designation
    valid_data = atlas_data[~atlas_data['designation'].isna() & (atlas_data['designation'] != '')]
    print(f"Filtered to {len(valid_data)} entries with non-empty designation values.")
    
    # Find duplicates based on 'number' and 'designation'
    duplicate_mask = valid_data.duplicated(subset=['number', 'designation'], keep=False)
    duplicates = valid_data[duplicate_mask].sort_values(by=['number', 'designation'])
    
    # Count duplicates
    duplicate_count = len(duplicates)
    unique_duplicate_pairs = duplicates[['number', 'designation']].drop_duplicates().shape[0]
    
    # Save duplicates to CSV
    if duplicate_count > 0:
        duplicates.to_csv("duplicates.csv", sep=';', index=False)
        print(f"Found {duplicate_count} duplicate entries ({unique_duplicate_pairs} unique number/designation pairs).")
        print(f"Duplicates saved to duplicates.csv")
        
        # Show sample of duplicates
        sample_size = min(5, unique_duplicate_pairs)
        if sample_size > 0:
            print("\nSample of duplicates:")
            sample_pairs = duplicates[['number', 'designation']].drop_duplicates().head(sample_size)
            
            for _, pair in sample_pairs.iterrows():
                number, designation = pair['number'], pair['designation']
                matches = duplicates[(duplicates['number'] == number) & 
                                     (duplicates['designation'] == designation)]
                
                print(f"\nNumber: {number}, Designation: {designation}")
                print(f"Found {len(matches)} entries with these values:")
                
                for i, (_, entry) in enumerate(matches.iterrows(), 1):
                    print(f"  {i}. SLOID: {entry['sloid']}, Description: {entry['designationOfficial']}")
    else:
        print("No duplicates found.")
    
    return duplicate_count, total_entries

if __name__ == "__main__":
    duplicate_count, total_entries = find_duplicates()
    print(f"\nSummary: {duplicate_count} duplicated entries out of {total_entries} total entries "
          f"({duplicate_count/total_entries:.2%})")
