"""
This files creates to csv files, one with all the ATLAS stops and one with route info for each stop.
"""
#import ATLAS data
import requests
import zipfile
import io
import csv
import pandas as pd
import os
import tqdm

def get_atlas_stops(atlas_stops_csv_output_path, download_url):

    response = requests.get(download_url)
    if response.status_code == 200:
        print("Download successful, extracting ZIP file...")
        with zipfile.ZipFile(io.BytesIO(response.content)) as z:
            # List all files in the ZIP archive
            csv_files = z.namelist()
            print("Files in the ZIP:", csv_files)
            
            if csv_files:
                # Assume the CSV file is the first one in the archive
                csv_filename = csv_files[0]
                print("Extracting:", csv_filename)

                # Extract the CSV file and process it before saving
                with z.open(csv_filename) as f:
                    # Read the CSV data
                    csv_data = pd.read_csv(f, sep=";")
                    
                    # Filter for Switzerland (country code 85)
                    csv_data = csv_data[csv_data['uicCountryCode'] == 85]
                    
                    # Identify rows without coordinates
                    no_coords = csv_data[csv_data[['wgs84North', 'wgs84East']].isna().any(axis=1)]
                    
                    # Find boarding platforms
                    boarding_platforms = csv_data[csv_data['trafficPointElementType'] == 'BOARDING_PLATFORM']
                    print(f"Found {len(boarding_platforms)} BOARDING_PLATFORM entries in ATLAS data.")
                    
                    # Remove rows without coordinates
                    csv_data = csv_data.dropna(subset=['wgs84North', 'wgs84East'])
                    
                    print(f"ATLAS data processed: {len(csv_data)} rows with coordinates, {len(no_coords)} without.")
                    
                    # Save the processed data
                    csv_data.to_csv(atlas_stops_csv_output_path, sep=";", index=False)
                    
                print(f"Processed CSV file saved as: {atlas_stops_csv_output_path}")
            else:
                print("No CSV file found in the ZIP archive.")
    else:
        print("Failed to download file. Status code:", response.status_code)

def extract_route_direction_per_stop(gtfs_data):
    """Extracts route and direction information for each stop from GTFS data."""
    stop_times_with_trip = pd.merge(gtfs_data['stop_times'], gtfs_data['trips'][['trip_id', 'route_id', 'direction_id']], on='trip_id', how='inner')

    # First get unique stop-route-direction combinations
    stop_route_direction = stop_times_with_trip[['stop_id', 'route_id', 'direction_id']].drop_duplicates()

    # Then join with routes to get route names
    route_cols = ['route_id', 'route_short_name', 'route_long_name']
    stop_route_direction = pd.merge(
        stop_route_direction, 
        gtfs_data['routes'][route_cols],
        on='route_id', 
        how='left'
    )
    return stop_route_direction

def match_gtfs_to_atlas(gtfs_data, traffic_points):
    """
    Match GTFS stops to ATLAS stops based on UIC number and local reference.
    
    Args:
        gtfs_data: Dictionary with GTFS dataframes
        traffic_points: DataFrame with ATLAS stops data
        
    Returns:
        DataFrame with matched stops (stop_id to sloid mapping)
    """
    print("\nMatching GTFS stops to ATLAS stops...")
    
    # 1. Filter for GTFS stops whose stop_id starts with "85"
    gtfs_stops = gtfs_data['stops'][gtfs_data['stops']['stop_id'].str.startswith('85')].copy()
    
    # 2. Parse GTFS stop_ids to extract uic_number and local_ref
    def parse_stop_id(stop_id):
        try:
            parts = stop_id.split(':')
            
            # Always take first part as UIC number
            uic_number = parts[0]
            
            # If there are at least 3 parts, take the third as local_ref
            if len(parts) >= 3:
                local_ref = parts[2]
                return uic_number, local_ref
            # Otherwise just return UIC with no local_ref
            return uic_number, None
        except:
            # Add to review file for manual inspection
            with open("review_stop_ids.txt", "a") as f:
                f.write(f"{stop_id}\n")
            return stop_id, None
            
    # Apply parsing function and create normalized local_ref
    gtfs_stops['uic_number'], gtfs_stops['local_ref'] = zip(*gtfs_stops['stop_id'].apply(parse_stop_id))
    gtfs_stops['uic_number'] = gtfs_stops['uic_number'].astype(str)
    
    # 3. Normalize local_ref (10000->1, 10001->2)
    def normalize_local_ref(ref):
        if not pd.isna(ref):
            if ref == '10000': return '1'
            elif ref == '10001': return '2'
        return ref
        
    gtfs_stops['normalized_local_ref'] = gtfs_stops['local_ref'].apply(normalize_local_ref)
    
    # 4. Prepare ATLAS entries
    traffic_points_copy = traffic_points.copy()
    traffic_points_copy['number'] = traffic_points_copy['number'].astype(str)
    
    # 5. Create indexes for faster lookups
    uic_to_atlas = {}
    for number, group in traffic_points_copy.groupby('number'):
        uic_to_atlas[number] = group.to_dict('records')
    
    # 6. Match GTFS stops to ATLAS stops
    matches = []
    
    # First do simple matching using merge
    print("Performing simple matches...")
    simple_matches = pd.merge(
        gtfs_stops[['stop_id', 'uic_number', 'normalized_local_ref']],
        traffic_points_copy[['sloid', 'number', 'designation']],
        left_on=['uic_number', 'normalized_local_ref'],
        right_on=['number', 'designation'],
        how='inner'
    )[['stop_id', 'sloid']]
    
    # Get remaining unmatched stops for advanced matching
    matched_stop_ids = set(simple_matches['stop_id'])
    unmatched_gtfs = gtfs_stops[~gtfs_stops['stop_id'].isin(matched_stop_ids)]
    
    print(f"Found {len(simple_matches)} simple matches")
    
    return simple_matches

def create_route_stops_mapping(integrated_data):
    """
    Create a DataFrame with one row per route+direction, listing all stops that belong to that route.
    
    Args:
        integrated_data: DataFrame with stop_id, sloid, route_id, direction_id, etc.
        
    Returns:
        DataFrame with route_id, direction_id, route_short_name, route_long_name, stops_list columns
    """
    print("\nCreating route to stops mapping...")
    
    # Group by route_id and direction_id
    route_groups = integrated_data.groupby(['route_id', 'direction_id'])
    
    # Create rows with route info and list of stops
    routes_with_stops = []
    
    for (route_id, direction_id), group in route_groups:
        # Get route info from first row (should be same for all rows in group)
        first_row = group.iloc[0]
        route_short_name = first_row['route_short_name']
        route_long_name = first_row['route_long_name']
        
        # Get list of sloids (stops) for this route+direction
        stops_list = group['sloid'].dropna().tolist()
        
        # Create a row for this route+direction combination
        route_row = {
            'route_id': route_id,
            'direction_id': direction_id,
            'route_short_name': route_short_name,
            'route_long_name': route_long_name,
            'stops_count': len(stops_list),
            'stops_list': stops_list
        }
        
        routes_with_stops.append(route_row)
    
    # Convert to DataFrame
    routes_df = pd.DataFrame(routes_with_stops)
    
    # Add stop_list as JSON string column for CSV export
    routes_df['stops_json'] = routes_df['stops_list'].apply(lambda x: pd.Series(x).to_json(orient='values'))
    
    print(f"Created mapping for {len(routes_df)} route+direction combinations")
    return routes_df


if __name__ == "__main__":
    atlas_stops_csv_output_path="stops_ATLAS.csv"
    # Check info here: https://data.opentransportdata.swiss/en/dataset/traffic-points-actual-date
    download_url = "https://data.opentransportdata.swiss/en/dataset/traffic-points-actual-date/permalink"
    get_atlas_stops(atlas_stops_csv_output_path, download_url)
    
    gtfs_folder = "gtfs_fp2024_2024-09-30"  # Folder containing GTFS files

    # Load relevant GTFS files
    gtfs_data = {file: pd.read_csv(f"{gtfs_folder}/{file}.txt", low_memory=False) for file in ['routes', 'trips', 'stops', 'stop_times']}
    
    # Load traffic points data
    traffic_points = pd.read_csv(atlas_stops_csv_output_path, sep=';')
    
    # Extract route and direction information per stop
    route_direction_info = extract_route_direction_per_stop(gtfs_data)
    
    # Match GTFS stops to ATLAS stops
    matches = match_gtfs_to_atlas(gtfs_data, traffic_points)
    
    # Merge matches with GTFS stops to get full stop information
    linked_stops = pd.merge(gtfs_data['stops'], matches, on='stop_id', how='left')
    
    if linked_stops is not None and route_direction_info is not None:
        try:
            integrated_data = pd.merge(linked_stops, route_direction_info, on='stop_id', how='inner')
            
            # Save route-stops mapping to CSV
            route_stops_mapping = create_route_stops_mapping(integrated_data)
            route_stops_mapping.to_csv("atlas_routes_with_stops.csv", index=False)
            
            # Save original stop-routes mapping as well
            (integrated_data[['stop_id', 'sloid', 'route_id', 'route_short_name', 
                           'route_long_name', 'direction_id']]
            .sort_values(by='sloid')
            .to_csv("atlas_routes.csv", index=False))
            
            # Generate matching statistics
            total_gtfs_stops = len(gtfs_data['stops'])
            total_atlas_stops = len(traffic_points)
            matched_stops = linked_stops['sloid'].notna().sum()
            unique_sloids_matched = integrated_data['sloid'].dropna().nunique()
            
            print("\n=== Matching Statistics ===")
            print(f"Total GTFS stops: {total_gtfs_stops}")
            print(f"Total ATLAS stops: {total_atlas_stops}")
            print(f"GTFS stops matched to ATLAS sloids: {matched_stops} ({matched_stops/total_gtfs_stops:.1%})")
            print(f"Unique ATLAS sloids matched to routes: {unique_sloids_matched} ({unique_sloids_matched/total_atlas_stops:.1%})")
            print(f"Saved {len(integrated_data)} rows to atlas_routes.csv")
            print(f"Saved {len(route_stops_mapping)} rows to atlas_routes_with_stops.csv")
            print("===========================")
            
            # Analyze unmatched entries
            print("\n=== Unmatched Entries Analysis ===")
            
            # Find ATLAS entries that weren't matched
            matched_sloids = set(integrated_data['sloid'].dropna())
            unmatched_atlas = traffic_points[~traffic_points['sloid'].isin(matched_sloids)]
            
            # Find GTFS stops that weren't matched to ATLAS stops
            matched_stops_ids = set(linked_stops[linked_stops['sloid'].notna()]['stop_id'])
            gtfs_stops = gtfs_data['stops'][gtfs_data['stops']['stop_id'].str.startswith('85')].copy()
            unmatched_gtfs = gtfs_stops[~gtfs_stops['stop_id'].isin(matched_stops_ids)]
            
            # Print sample of unmatched ATLAS entries
            print(f"\nTotal unmatched ATLAS entries: {len(unmatched_atlas)} ({len(unmatched_atlas)/len(traffic_points):.1%} of all ATLAS entries)")
            if not unmatched_atlas.empty:
                print("\n10 sample unmatched ATLAS entries:")
                sample_atlas = unmatched_atlas.sample(min(10, len(unmatched_atlas)))
                for _, row in sample_atlas.iterrows():
                    print(f"SLOID: {row['sloid']}, Designation Official: {row['designationOfficial']}, Number: {row['number']}, " + 
                          f"Designation: {row['designation']}")
            
            # Print sample of unmatched GTFS entries
            print(f"\nTotal unmatched GTFS stops: {len(unmatched_gtfs)} ({len(unmatched_gtfs)/len(gtfs_data['stops']):.1%} of all GTFS stops)")
            if not unmatched_gtfs.empty:
                print("\n10 sample unmatched GTFS stops:")
                sample_gtfs = unmatched_gtfs.sample(min(10, len(unmatched_gtfs)))
                for _, row in sample_gtfs.iterrows():
                    print(f"Stop ID: {row['stop_id']}, Stop Name: {row['stop_name']}")
            
            print("===========================")
            
            # After the matching statistics, add this code to identify duplicates
            duplicates = matches[matches.duplicated(subset=['sloid'], keep=False)]
            duplicate_sloids = duplicates['sloid'].unique()
            print(f"\nFound {len(duplicate_sloids)} ATLAS sloids matched to multiple GTFS stops:")
            for sloid in duplicate_sloids:
                matching_stops = matches[matches['sloid'] == sloid]['stop_id'].tolist()
                print(f"SLOID {sloid} matched to {len(matching_stops)} GTFS stops: {matching_stops}")
        except Exception as e:
            print(f"Error processing data: {e}")
            import traceback
            traceback.print_exc()
    
    print("Done!")

