from sqlalchemy import create_engine, Column, Integer, String, Float, Text, JSON
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
import json, math
import pandas as pd
import numpy as np
from matching_script import final_pipeline


# Database Setup
DATABASE_URI = 'mysql+pymysql://root:@localhost/stops_db'
engine = create_engine(DATABASE_URI)
Session = sessionmaker(bind=engine)
session = Session()
Base = declarative_base()

def sanitize_for_json(obj):
    if isinstance(obj, dict):
        return {k: sanitize_for_json(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [sanitize_for_json(item) for item in obj]
    elif isinstance(obj, float):
        return None if math.isnan(obj) else obj
    else:
        return obj

def safe_value(val, default=None):
    """Safely handle NaN, None, and other problematic values for MySQL"""
    if val is None:
        return default
    if isinstance(val, float) and (math.isnan(val) or math.isinf(val)):
        return default
    if pd.isna(val):
        return default
    return val

# Stop Model
class Stop(Base):
    __tablename__ = 'stops'
    id = Column(Integer, primary_key=True)
    sloid = Column(String(100))  # ATLAS stop identifier
    stop_type = Column(String(50))   # 'matched', 'station', 'unmatched', 'osm'
    match_type = Column(String(50))  # e.g. 'exact', 'name', 'distance_matching_1'
    
    # ATLAS attributes
    atlas_designation = Column(String(255))           # ATLAS designation (from CSV)
    atlas_designation_official = Column(String(255))  # Official designation (from CSV or fallback)
    atlas_lat = Column(Float)                         # ATLAS latitude (wgs84North)
    atlas_lon = Column(Float)                         # ATLAS longitude (wgs84East)
    atlas_business_org_abbr = Column(String(100))     # ATLAS servicePointBusinessOrganisationAbbreviationEn
    
    # UIC Reference (merged field from ATLAS number and OSM uic_ref)
    uic_ref = Column(String(100), index=True)         # UIC reference from either source
    
    # OSM attributes
    osm_node_id = Column(String(100))   # OSM node identifier
    osm_local_ref = Column(String(100)) # OSM "local_ref" tag
    osm_name = Column(String(255))      # OSM stop name (for display)
    osm_uic_name = Column(String(255))  # OSM "uic_name" attribute, if available
    osm_network = Column(String(255))   # OSM "network" tag
    osm_operator = Column(String(255))  # OSM "operator" tag
    osm_public_transport = Column(String(255))  # Add this line for public_transport tag
    osm_lat = Column(Float)             # OSM latitude (if applicable)
    osm_lon = Column(Float)             # OSM longitude (if applicable)
    
    distance_m = Column(Float)  # Distance between ATLAS and OSM coordinates (if matched)
    
    # Route information as JSON
    routes_atlas = Column(JSON)  # List of {route_id, direction_id, route_short_name} for ATLAS
    routes_osm = Column(JSON)    # List of {route_id, direction_id, route_short_name} for OSM

# Route and Direction Model
class RouteAndDirection(Base):
    __tablename__ = 'routes_and_directions'
    id = Column(Integer, primary_key=True)
    direction_id = Column(String(20))
    osm_route_id = Column(String(100))
    osm_nodes_json = Column(JSON)       # List of node_ids that belong to this OSM route
    atlas_route_id = Column(String(100))
    atlas_sloids_json = Column(JSON)    # List of sloids that belong to this ATLAS route
    route_name = Column(String(255))
    route_short_name = Column(String(50))
    route_long_name = Column(String(255))
    route_type = Column(String(50))
    match_type = Column(String(50))     # 'matched', 'osm_only', 'atlas_only'

# Create (or refresh) the tables.
Base.metadata.create_all(engine)

# --------------------------
# Helper Functions for Data Import
# --------------------------
def validate_coordinates(rec, lat_field, lon_field, id_field, id_value, rec_type):
    """Validate and convert coordinates, returning (lat, lon) or (None, None) if invalid"""
    lat_val = safe_value(rec.get(lat_field))
    lon_val = safe_value(rec.get(lon_field))
    
    if lat_val is None or lon_val is None:
        print(f"Warning: Skipping {rec_type} record with {id_field} {id_value} due to missing coordinates.")
        return None, None
    
    try:
        lat = float(lat_val)
        lon = float(lon_val)
        # Check for NaN after conversion
        if math.isnan(lat) or math.isnan(lon):
            print(f"Warning: Skipping {rec_type} record with {id_field} {id_value} due to NaN coordinates.")
            return None, None
        return lat, lon
    except Exception as e:
        print(f"Warning: Skipping {rec_type} record with {id_field} {id_value} due to conversion error: {e}")
        return None, None

def get_from_tags(rec, tag_name, default=None):
    """Safely extract a tag from the tags dictionary"""
    tags = rec.get('tags', {})
    if not tags:
        return default
    return safe_value(tags.get(tag_name), default)

def create_stop_record(
    stop_type, match_type=None,
    sloid=None, uic_ref=None, atlas_designation=None, atlas_designation_official=None,
    atlas_business_org_abbr=None, atlas_lat=None, atlas_lon=None,
    osm_node_id=None, osm_local_ref=None, osm_name=None, osm_uic_name=None,
    osm_network=None, osm_operator=None, osm_public_transport=None, osm_lat=None, osm_lon=None,
    distance_m=None, routes_atlas=None, routes_osm=None
):
    """Create a Stop record with default values for all fields"""
    return Stop(
        sloid=sloid,
        stop_type=stop_type,
        match_type=match_type,
        atlas_designation=atlas_designation,
        atlas_designation_official=atlas_designation_official,
        atlas_business_org_abbr=atlas_business_org_abbr,
        atlas_lat=atlas_lat,
        atlas_lon=atlas_lon,
        uic_ref=uic_ref,
        osm_node_id=osm_node_id,
        osm_local_ref=osm_local_ref,
        osm_name=osm_name,
        osm_uic_name=osm_uic_name,
        osm_network=osm_network,
        osm_operator=osm_operator,
        osm_public_transport=osm_public_transport,
        osm_lat=osm_lat,
        osm_lon=osm_lon,
        distance_m=distance_m,
        routes_atlas=routes_atlas,
        routes_osm=routes_osm
    )

def load_route_data():
    """Load route data from CSV files and create mappings for stops to routes"""
    # Load ATLAS routes
    atlas_routes_mapping = {}
    try:
        atlas_routes_df = pd.read_csv("atlas_routes.csv")
        for _, row in atlas_routes_df.iterrows():
            sloid = row['sloid']
            if pd.notna(sloid):
                route_info = {
                    'route_id': row['route_id'],
                    'direction_id': str(row['direction_id']),
                    'route_short_name': row['route_short_name'] if pd.notna(row['route_short_name']) else None,
                    'route_long_name': row['route_long_name'] if pd.notna(row['route_long_name']) else None
                }
                
                if sloid not in atlas_routes_mapping:
                    atlas_routes_mapping[sloid] = []
                atlas_routes_mapping[sloid].append(route_info)
        print(f"Loaded route information for {len(atlas_routes_mapping)} ATLAS stops")
    except Exception as e:
        print(f"Error loading ATLAS routes: {e}")
        
    # Load OSM routes
    osm_routes_mapping = {}
    try:
        osm_routes_df = pd.read_csv("osm_nodes_with_routes.csv")
        for _, row in osm_routes_df.iterrows():
            node_id = str(row['node_id'])
            if pd.notna(node_id):
                route_info = {
                    'route_id': row['gtfs_route_id'] if pd.notna(row['gtfs_route_id']) else None,
                    'direction_id': str(int(float(row['direction_id']))) if pd.notna(row['direction_id']) else None,
                    'route_name': row['route_name'] if pd.notna(row['route_name']) else None
                }
                
                if node_id not in osm_routes_mapping:
                    osm_routes_mapping[node_id] = []
                osm_routes_mapping[node_id].append(route_info)
        print(f"Loaded route information for {len(osm_routes_mapping)} OSM nodes")
    except Exception as e:
        print(f"Error loading OSM routes: {e}")
        
    return atlas_routes_mapping, osm_routes_mapping

def build_route_direction_mapping():
    """Build mappings for routes and directions"""
    # Maps for route+direction to nodes
    osm_route_dir_to_nodes = {}
    atlas_route_dir_to_sloids = {}
    
    try:
        # Process OSM routes
        osm_routes_df = pd.read_csv("osm_nodes_with_routes.csv")
        for _, row in osm_routes_df.iterrows():
            if pd.notna(row['gtfs_route_id']) and pd.notna(row['direction_id']):
                route_id = row['gtfs_route_id']
                direction_id = str(int(float(row['direction_id'])))
                node_id = str(row['node_id'])
                
                key = (route_id, direction_id)
                if key not in osm_route_dir_to_nodes:
                    osm_route_dir_to_nodes[key] = {
                        'nodes': [],
                        'route_name': row['route_name'] if pd.notna(row['route_name']) else None
                    }
                osm_route_dir_to_nodes[key]['nodes'].append(node_id)
        
        # Process ATLAS routes
        atlas_routes_df = pd.read_csv("atlas_routes.csv")
        for _, row in atlas_routes_df.iterrows():
            if pd.notna(row['route_id']) and pd.notna(row['direction_id']) and pd.notna(row['sloid']):
                route_id = row['route_id']
                direction_id = str(row['direction_id'])
                sloid = row['sloid']
                
                key = (route_id, direction_id)
                if key not in atlas_route_dir_to_sloids:
                    atlas_route_dir_to_sloids[key] = {
                        'sloids': [],
                        'route_short_name': row['route_short_name'] if pd.notna(row['route_short_name']) else None,
                        'route_long_name': row['route_long_name'] if pd.notna(row['route_long_name']) else None
                    }
                atlas_route_dir_to_sloids[key]['sloids'].append(sloid)
        
        print(f"Built route+direction to nodes mapping for {len(osm_route_dir_to_nodes)} OSM routes")
        print(f"Built route+direction to sloids mapping for {len(atlas_route_dir_to_sloids)} ATLAS routes")
    except Exception as e:
        print(f"Error building route-direction mappings: {e}")
        
    return osm_route_dir_to_nodes, atlas_route_dir_to_sloids

# --------------------------
# Data Import Function
# --------------------------
def import_to_database(base_data):
    """
    Fully refresh the database, inserting:
      - Matched stops (with both ATLAS and OSM attributes)
      - Unmatched ATLAS stops (ensuring designation_official is present)
      - Unmatched OSM nodes
      - Route and direction information
    """
    print("Deleting existing data from database...")
    session.query(RouteAndDirection).delete()
    session.query(Stop).delete()
    session.commit()
    print("Existing data deleted. Starting new import.")
    
    # Load route information
    atlas_routes_mapping, osm_routes_mapping = load_route_data()
    
    # Build route+direction to nodes/sloids mappings for routes_and_directions table
    osm_route_dir_to_nodes, atlas_route_dir_to_sloids = build_route_direction_mapping()
    
    # --- Insert Matched Records ---
    matched_records = base_data.get('matched', [])
    for rec in matched_records:
        atlas_lat, atlas_lon = validate_coordinates(
            rec, 'csv_lat', 'csv_lon', 'sloid', rec.get('sloid'), 'matched'
        )
        if atlas_lat is None:  # Skip if coordinates are invalid
            continue
        
        # OSM coordinates (if available)
        try:
            osm_lat = float(safe_value(rec.get('osm_lat'))) if safe_value(rec.get('osm_lat')) is not None else None
            osm_lon = float(safe_value(rec.get('osm_lon'))) if safe_value(rec.get('osm_lon')) is not None else None
            # Check for NaN after conversion
            if osm_lat is not None and math.isnan(osm_lat):
                osm_lat = None
            if osm_lon is not None and math.isnan(osm_lon):
                osm_lon = None
        except Exception:
            osm_lat = None
            osm_lon = None
        
        # Determine designation; fallback from csv_designation_official then csv_designation
        designation_official = safe_value(rec.get('csv_designation_official')) or safe_value(rec.get('designationOfficial')) or safe_value(rec.get('csv_designation')) or ""
        
        # Use the ATLAS number as the uic_ref (OSM uic_ref would be the same in a match)
        uic_ref = safe_value(rec.get('number'), "")
        
        # Get the business organization abbreviation and OSM network/operator
        business_org_abbr = safe_value(rec.get('csv_business_org_abbr', ''))
        osm_network = safe_value(rec.get('osm_network', ''))
        osm_operator = safe_value(rec.get('osm_operator', ''))
        
        # Get route information
        sloid = safe_value(rec.get('sloid'))
        osm_node_id = safe_value(rec.get('osm_node_id'))
        
        routes_atlas_data = atlas_routes_mapping.get(sloid, []) if sloid else []
        routes_osm_data = osm_routes_mapping.get(osm_node_id, []) if osm_node_id else []
        
        stop = create_stop_record(
            stop_type='matched',
            match_type=safe_value(rec.get('match_type')),
            sloid=sloid,
            uic_ref=uic_ref,
            atlas_designation=safe_value(rec.get('csv_designation'), ""),
            atlas_designation_official=designation_official,
            atlas_business_org_abbr=business_org_abbr,
            atlas_lat=atlas_lat,
            atlas_lon=atlas_lon,
            osm_node_id=osm_node_id,
            osm_local_ref=safe_value(rec.get('osm_local_ref')),
            osm_name=safe_value(rec.get('osm_name')) or get_from_tags(rec, 'name'),
            osm_uic_name=safe_value(rec.get('osm_uic_name')) or get_from_tags(rec, 'uic_name'),
            osm_network=osm_network,
            osm_operator=osm_operator,
            osm_public_transport=safe_value(rec.get('osm_public_transport')),
            osm_lat=osm_lat,
            osm_lon=osm_lon,
            distance_m=safe_value(rec.get('distance_m')),
            routes_atlas=routes_atlas_data if routes_atlas_data else None,
            routes_osm=routes_osm_data if routes_osm_data else None
        )
        session.add(stop)
    
    # --- Insert Unmatched ATLAS Records ---
    unmatched_records = base_data.get('unmatched_atlas', [])
    for rec in unmatched_records:
        atlas_lat, atlas_lon = validate_coordinates(
            rec, 'wgs84North', 'wgs84East', 'sloid', rec.get('sloid'), 'unmatched ATLAS'
        )
        if atlas_lat is None:  # Skip if coordinates are invalid
            continue
        
        # Ensure designation_official is always present (fallback to designation if needed)
        designation_official = safe_value(rec.get('designationOfficial')) or safe_value(rec.get('designation')) or ""
        uic_ref = safe_value(rec.get('number'), "")
        business_org_abbr = safe_value(rec.get('servicePointBusinessOrganisationAbbreviationEn', ''))
        
        # Get route information for this ATLAS stop
        sloid = safe_value(rec.get('sloid'))
        routes_atlas_data = atlas_routes_mapping.get(sloid, []) if sloid else []
        
        stop = create_stop_record(
            stop_type='unmatched',
            sloid=sloid,
            uic_ref=uic_ref,
            atlas_designation=safe_value(rec.get('designation'), ""),
            atlas_designation_official=designation_official,
            atlas_business_org_abbr=business_org_abbr,
            atlas_lat=atlas_lat,
            atlas_lon=atlas_lon,
            routes_atlas=routes_atlas_data if routes_atlas_data else None
        )
        session.add(stop)
    
    # --- Insert Unmatched OSM Records ---
    unmatched_osm_records = base_data.get('unmatched_osm', [])
    for rec in unmatched_osm_records:
        osm_lat, osm_lon = validate_coordinates(
            rec, 'lat', 'lon', 'node_id', rec.get('node_id'), 'unmatched OSM'
        )
        if osm_lat is None:  # Skip if coordinates are invalid
            continue
        
        # Get OSM network and operator
        osm_network = get_from_tags(rec, 'network', '')
        osm_operator = get_from_tags(rec, 'operator', '')
        
        # Get OSM uic_ref if available
        uic_ref = get_from_tags(rec, 'uic_ref', '')
        
        # Get route information for this OSM node
        osm_node_id = str(safe_value(rec.get('node_id')))
        routes_osm_data = osm_routes_mapping.get(osm_node_id, []) if osm_node_id else []
        
        stop = create_stop_record(
            stop_type='osm',
            uic_ref=uic_ref,
            osm_node_id=osm_node_id,
            osm_local_ref=safe_value(rec.get('local_ref')),
            osm_name=safe_value(rec.get('name')) or get_from_tags(rec, 'name'),
            osm_uic_name=get_from_tags(rec, 'uic_name'),
            osm_network=osm_network,
            osm_operator=osm_operator,
            osm_public_transport=get_from_tags(rec, 'public_transport', ''),
            osm_lat=osm_lat,
            osm_lon=osm_lon,
            routes_osm=routes_osm_data if routes_osm_data else None
        )
        session.add(stop)
    
    # --- Insert Route and Direction Records ---
    # Find matches between OSM and ATLAS routes using route_id and direction_id
    matched_routes = 0
    osm_only_routes = 0
    atlas_only_routes = 0
    
    # Process all OSM routes first (including matching to ATLAS)
    for (osm_route_id, direction_id), osm_data in osm_route_dir_to_nodes.items():
        # Check if this route+direction combination exists in ATLAS
        atlas_data = atlas_route_dir_to_sloids.get((osm_route_id, direction_id))
        
        if atlas_data:
            # This is a matched route
            route_record = RouteAndDirection(
                direction_id=direction_id,
                osm_route_id=osm_route_id,
                osm_nodes_json=osm_data['nodes'],
                atlas_route_id=osm_route_id,  # Same ID for matched routes
                atlas_sloids_json=atlas_data['sloids'],
                route_name=osm_data['route_name'],
                route_short_name=atlas_data['route_short_name'],
                route_long_name=atlas_data['route_long_name'],
                route_type=None,  # Could be added if available
                match_type='matched'
            )
            matched_routes += 1
        else:
            # This is an OSM-only route
            route_record = RouteAndDirection(
                direction_id=direction_id,
                osm_route_id=osm_route_id,
                osm_nodes_json=osm_data['nodes'],
                atlas_route_id=None,
                atlas_sloids_json=None,
                route_name=osm_data['route_name'],
                route_short_name=None,
                route_long_name=None,
                route_type=None,
                match_type='osm_only'
            )
            osm_only_routes += 1
            
        session.add(route_record)
    
    # Process ATLAS-only routes
    processed_keys = set(osm_route_dir_to_nodes.keys())
    for (atlas_route_id, direction_id), atlas_data in atlas_route_dir_to_sloids.items():
        # Skip if we already processed this route+direction combination
        if (atlas_route_id, direction_id) in processed_keys:
            continue
            
        # This is an ATLAS-only route
        route_record = RouteAndDirection(
            direction_id=direction_id,
            osm_route_id=None,
            osm_nodes_json=None,
            atlas_route_id=atlas_route_id,
            atlas_sloids_json=atlas_data['sloids'],
            route_name=None,
            route_short_name=atlas_data['route_short_name'],
            route_long_name=atlas_data['route_long_name'],
            route_type=None,
            match_type='atlas_only'
        )
        atlas_only_routes += 1
        session.add(route_record)
    
    session.commit()
    print(f"Route statistics: {matched_routes} matched, {osm_only_routes} OSM-only, {atlas_only_routes} ATLAS-only")
    session.close()
    print("Data import complete!")

if __name__ == "__main__":
    # Run the final pipeline to obtain base_data in-memory
    print("Running the final pipeline to obtain base data...")
    base_data = final_pipeline()
    # Directly import the in-memory base_data into the database
    print("Importing data into the database...")
    import_to_database(base_data)
    print("Process completed successfully!")