import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from collections import defaultdict
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base


# Database Setup
DATABASE_URI = 'mysql+pymysql://root:@localhost/stops_db'
engine = create_engine(DATABASE_URI)
Session = sessionmaker(bind=engine)
session = Session()
Base = declarative_base()


def ensure_plots_directory():
    """Create the plots directory if it doesn't exist."""
    if not os.path.exists('plots'):
        os.makedirs('plots')


def plot_distance_distributions(all_matches_df):
    """
    Plot distance distributions for each matching method.
    
    Args:
        all_matches_df: DataFrame containing all the matches
    """
    ensure_plots_directory()
    
    # Group by match_type for analysis
    match_types = all_matches_df['match_type'].unique()
    
    # Create a single figure with subplots for distance distribution by match type
    plt.figure(figsize=(15, 10))
    
    # Plot a histogram of distances by match type using Seaborn
    sns.histplot(data=all_matches_df, x='distance_m', hue='match_type', multiple='stack')
    plt.title('Distance Distribution by Match Type (All Methods)')
    plt.xlabel('Distance (meters)')
    plt.ylabel('Count')
    plt.grid(True, alpha=0.3)
    plt.xlim(0, 100)  # Set x-axis limit to 100m for the combined plot
    plt.savefig('plots/distance_distribution_all.png')
    plt.close()
    
    # Create individual histograms for each match type
    for match_type in match_types:
        subset = all_matches_df[all_matches_df['match_type'] == match_type]
        
        plt.figure(figsize=(10, 6))
        sns.histplot(data=subset, x='distance_m', kde=True)
        plt.title(f'Distance Distribution: {match_type}')
        plt.xlabel('Distance (meters)')
        plt.ylabel('Count')
        plt.grid(True, alpha=0.3)
        
        # Set different x-axis limits based on match type
        if 'distance' in match_type.lower():
            plt.xlim(0, 50)  # Set x-axis limit to 50m for distance matching methods
        else:
            plt.xlim(0, 100)  # Set x-axis limit to 100m for other methods
            
        plt.savefig(f'plots/distance_{match_type}.png')
        plt.close()


def get_top_distances(all_matches_df, output_file="plots/top_distances.txt"):
    """
    Extract and save the top 10 distances for each match type.
    
    Args:
        all_matches_df: DataFrame containing all the matches
        output_file: Path to save the results
    """
    ensure_plots_directory()
    
    # Group by match_type
    grouped_by_type = all_matches_df.groupby('match_type')
    
    # Open file for writing
    with open(output_file, 'w') as f:
        f.write("TOP 10 DISTANCES BY MATCHING METHOD\n")
        f.write("===================================\n\n")
        
        for match_type, group in grouped_by_type:
            # Sort by distance and get top 10
            top_distances = group.sort_values('distance_m', ascending=False).head(10)
            
            f.write(f"\n{match_type.upper()}\n")
            f.write("-" * len(match_type) + "\n")
            
            for idx, row in top_distances.iterrows():
                f.write(f"Distance: {row['distance_m']:.2f}m, SLOID: {row['sloid']}, ")
                f.write(f"OSM Node: {row['osm_node_id']}, ")
                if 'csv_designation' in row and pd.notna(row['csv_designation']):
                    f.write(f"Designation: {row['csv_designation']}, ")
                if 'osm_name' in row and pd.notna(row['osm_name']):
                    f.write(f"OSM Name: {row['osm_name']}")
                f.write("\n")
            
            # Summary statistics
            f.write(f"\nSummary Statistics for {match_type}:\n")
            f.write(f"Count: {len(group)}\n")
            f.write(f"Min: {group['distance_m'].min():.2f}m\n")
            f.write(f"Max: {group['distance_m'].max():.2f}m\n")
            f.write(f"Mean: {group['distance_m'].mean():.2f}m\n")
            f.write(f"Median: {group['distance_m'].median():.2f}m\n")
            f.write("\n" + "=" * 50 + "\n")
        
        # Add global statistics
        f.write("\nGLOBAL STATISTICS (ALL MATCHING METHODS)\n")
        f.write("=======================================\n")
        f.write(f"Total matches: {len(all_matches_df)}\n")
        f.write(f"Global maximum distance: {all_matches_df['distance_m'].max():.2f}m\n")
        f.write(f"Global average distance: {all_matches_df['distance_m'].mean():.2f}m\n")
        f.write(f"Global median distance: {all_matches_df['distance_m'].median():.2f}m\n")


def print_main_method_statistics(all_matches_df):
    """
    Calculates and prints the average and median distance for 
    each main matching method (distance, exact, name).
    
    Args:
        all_matches_df: DataFrame containing all the matches
    """
    # Create a new column that categorizes match_types into main methods
    all_matches_df['main_method'] = all_matches_df['match_type'].apply(
        lambda x: 'distance' if 'distance' in x.lower() else
                 ('exact' if 'exact' in x.lower() else
                 ('name' if 'name' in x.lower() else 'other'))
    )
    
    # Group by main_method
    grouped = all_matches_df.groupby('main_method')
    
    # Print the statistics for each main method
    print("\nSTATISTICS BY MAIN MATCHING METHOD")
    print("==================================")
    
    for method, group in grouped:
        if method != 'other':  # Exclude the 'other' category if you only want the three main methods
            print(f"\n{method.upper()}")
            print(f"Count: {len(group)}")
            print(f"Average: {group['distance_m'].mean():.2f}m")
            print(f"Median: {group['distance_m'].median():.2f}m")


def generate_distance_statistics():
    """
    Main function to generate all distance statistics and plots.
    Fetches data directly from the database.
    """
    print("Fetching match data from database...")
    
    # Query the database to get all matched stops with distance information
    from app import Stop  # Import the Stop model from app.py
    
    # Query all matched stops with distance information
    query = session.query(
        Stop.sloid, 
        Stop.osm_node_id, 
        Stop.atlas_designation.label('csv_designation'),
        Stop.osm_name,
        Stop.match_type,
        Stop.distance_m
    ).filter(
        Stop.stop_type == 'matched',
        Stop.distance_m != None
    )
    
    # Convert query results to DataFrame
    matches_df = pd.read_sql(query.statement, session.bind)
    
    # Filter out rows with missing distance values
    valid_matches = matches_df.dropna(subset=['distance_m'])
    
    # Generate plots
    plot_distance_distributions(valid_matches)
    
    # Get top distances
    get_top_distances(valid_matches)
    
    # Print statistics by main method
    print_main_method_statistics(valid_matches)
    
    print(f"Distance statistics complete. Plots saved to 'plots/' directory.")
    print(f"Top distances saved to 'plots/top_distances.txt'")


if __name__ == "__main__":
    # Run the distance statistics generation when the script is executed directly
    generate_distance_statistics()
