import xml.etree.ElementTree as ET
import csv
import pandas as pd
import matplotlib.pyplot as plt
from collections import defaultdict
import os

# Define the expected keys that determine if a node is "assigned"
expected_keys = {"uic_ref", "ref", "local_ref", "name", "network", "operator", "uic_name"}

# Overall counters
total_nodes = 0

# Counters for nodes with "uic_ref"
nodes_with_uic_ref = 0
uic_ref_with_ref = 0
uic_ref_with_local_ref = 0
uic_ref_with_both = 0
uic_ref_with_name = 0
uic_ref_with_network = 0
uic_ref_with_operator = 0
uic_ref_with_uic_name = 0

# Counter for nodes with all required tags
nodes_with_all_tags = 0

# Counters for nodes without "uic_ref"
nodes_without_uic_ref = 0
non_uic_ref_with_ref = 0
non_uic_ref_with_local_ref = 0
non_uic_ref_with_both = 0
non_uic_ref_with_name = 0
non_uic_ref_with_network = 0
non_uic_ref_with_operator = 0
non_uic_ref_with_uic_name = 0

# List to collect unassigned nodes (store the entire Element)
unassigned_node_elements = []

# Counter for unassigned nodes that have <tag k="aerialway" v="station" />
unassigned_aerialway_station = 0

# New counters for public transport platforms and stop positions
nodes_platform = 0
nodes_stop_position = 0
platforms_with_uic_ref = 0
stop_positions_with_uic_ref = 0
platforms_with_matching_stop = 0
platforms_without_matching_stop = 0

# Dictionary to store nodes with uic_ref for matching analysis
platform_uic_refs = set()
stop_position_uic_refs = set()

# Store examples of platform nodes with uic_ref but no matching stop_position
unmatched_platform_examples = []

# Parse the XML file
tree = ET.parse("osm_nodes.xml")
root = tree.getroot()

# First pass: collect all stop_position uic_refs
for node in root.findall(".//node"):
    tags = {tag.get("k"): tag.get("v") for tag in node.findall("tag")}
    if tags.get("public_transport") == "stop_position":
        if "uic_ref" in tags:
            stop_position_uic_refs.add(tags["uic_ref"])
            stop_positions_with_uic_ref += 1

# Iterate through each <node> element for main analysis
for node in root.findall(".//node"):
    total_nodes += 1
    node_id = node.get("id")
    node_lat = node.get("lat")
    node_lon = node.get("lon")
    
    # Build a dictionary of tag key-value pairs for this node
    tags = {tag.get("k"): tag.get("v") for tag in node.findall("tag")}
    
    # Check for public_transport=platform or public_transport=stop_position
    if tags.get("public_transport") == "platform":
        nodes_platform += 1
        if "uic_ref" in tags:
            platforms_with_uic_ref += 1
            platform_uic_refs.add(tags["uic_ref"])
            
            # Check if this platform has a matching stop_position
            if tags["uic_ref"] in stop_position_uic_refs:
                platforms_with_matching_stop += 1
            else:
                platforms_without_matching_stop += 1
                # Store example if we have fewer than 5
                if len(unmatched_platform_examples) < 5:
                    node_info = {
                        "id": node_id,
                        "lat": node_lat,
                        "lon": node_lon,
                        "uic_ref": tags.get("uic_ref", "N/A"),
                        "name": tags.get("name", "N/A")
                    }
                    unmatched_platform_examples.append(node_info)
    elif tags.get("public_transport") == "stop_position":
        nodes_stop_position += 1
    
    # Check if node contains any of the expected keys.
    if not any(key in tags for key in expected_keys):
        # This node is unassigned; store the entire node element.
        unassigned_node_elements.append(node)
        # Additionally, check if it has aerialway=station
        if tags.get("aerialway") == "station":
            unassigned_aerialway_station += 1
    
    # Check if node has all required tags
    required_tags = {"uic_ref", "local_ref", "name", "network", "operator", "uic_name"}
    if all(tag in tags for tag in required_tags):
        nodes_with_all_tags += 1
    
    # Process nodes with or without uic_ref
    if "uic_ref" in tags:
        nodes_with_uic_ref += 1
        if "ref" in tags:
            uic_ref_with_ref += 1
        if "local_ref" in tags:
            uic_ref_with_local_ref += 1
        if "ref" in tags and "local_ref" in tags:
            uic_ref_with_both += 1
        if "name" in tags:
            uic_ref_with_name += 1
        if "network" in tags:
            uic_ref_with_network += 1
        if "operator" in tags:
            uic_ref_with_operator += 1
        if "uic_name" in tags:
            uic_ref_with_uic_name += 1
    else:
        nodes_without_uic_ref += 1
        if "ref" in tags:
            non_uic_ref_with_ref += 1
        if "local_ref" in tags:
            non_uic_ref_with_local_ref += 1
        if "ref" in tags and "local_ref" in tags:
            non_uic_ref_with_both += 1
        if "name" in tags:
            non_uic_ref_with_name += 1
        if "network" in tags:
            non_uic_ref_with_network += 1
        if "operator" in tags:
            non_uic_ref_with_operator += 1
        if "uic_name" in tags:
            non_uic_ref_with_uic_name += 1

# Prepare the report data for CSV output
report_data = [
    ["Total nodes", total_nodes],
    ["Nodes with uic_ref", nodes_with_uic_ref],
    ["Nodes without uic_ref", nodes_without_uic_ref],
    ["Nodes with all these tags (uic_ref, local_ref, name, network, operator, uic_name)", nodes_with_all_tags],
    ["--- For nodes with uic_ref ---", ""],
    ["uic_ref nodes with ref", uic_ref_with_ref],
    ["uic_ref nodes with local_ref", uic_ref_with_local_ref],
    ["uic_ref nodes with both ref and local_ref", uic_ref_with_both],
    ["uic_ref nodes with name", uic_ref_with_name],
    ["uic_ref nodes with network", uic_ref_with_network],
    ["uic_ref nodes with operator", uic_ref_with_operator],
    ["uic_ref nodes with uic_name", uic_ref_with_uic_name],
    ["--- For nodes without uic_ref ---", ""],
    ["Non-uic_ref nodes with ref", non_uic_ref_with_ref],
    ["Non-uic_ref nodes with local_ref", non_uic_ref_with_local_ref],
    ["Non-uic_ref nodes with both ref and local_ref", non_uic_ref_with_both],
    ["Non-uic_ref nodes with name", non_uic_ref_with_name],
    ["Non-uic_ref nodes with network", non_uic_ref_with_network],
    ["Non-uic_ref nodes with operator", non_uic_ref_with_operator],
    ["Non-uic_ref nodes with uic_name", non_uic_ref_with_uic_name],
    ["--- Unassigned nodes ---", ""],
    ["Total unassigned nodes", len(unassigned_node_elements)],
    ["Unassigned nodes with aerialway=station", unassigned_aerialway_station],
    ["--- Public Transport Tags ---", ""],
    ["Nodes with public_transport=platform", nodes_platform],
    ["Nodes with public_transport=platform that have uic_ref", platforms_with_uic_ref],
    ["Nodes with public_transport=stop_position", nodes_stop_position],
    ["Nodes with public_transport=stop_position that have uic_ref", stop_positions_with_uic_ref],
    ["Platform nodes with matching stop_position (same uic_ref)", platforms_with_matching_stop],
    ["Platform nodes with uic_ref but no matching stop_position", platforms_without_matching_stop]
]

# Create a single output file for all text reports (combining CSV data with text reports)
txt_filename = "osm_analysis_report.txt"
with open(txt_filename, "w", encoding="utf-8") as txtfile:
    txtfile.write("OSM PUBLIC TRANSPORT DATA ANALYSIS\n")
    txtfile.write("================================\n\n")
    
    # Write tagging analysis data directly to the text report
    txtfile.write("TAGGING ANALYSIS\n")
    txtfile.write("---------------\n")
    for row in report_data:
        txtfile.write(f"{row[0]}: {row[1]}\n")
    txtfile.write("\n")
    
    # Write general statistics
    txtfile.write("GENERAL STATISTICS\n")
    txtfile.write("------------------\n")
    txtfile.write(f"Total nodes analyzed: {total_nodes}\n")
    txtfile.write(f"Nodes with uic_ref: {nodes_with_uic_ref}\n")
    txtfile.write(f"Nodes without uic_ref: {nodes_without_uic_ref}\n")
    txtfile.write(f"Nodes with all these tags (uic_ref, local_ref, name, network, operator, uic_name): {nodes_with_all_tags}\n\n")
    
    # Write public transport statistics
    txtfile.write("PUBLIC TRANSPORT STATISTICS\n")
    txtfile.write("--------------------------\n")
    txtfile.write(f"Total platform nodes: {nodes_platform}\n")
    txtfile.write(f"Platform nodes with uic_ref: {platforms_with_uic_ref}\n")
    txtfile.write(f"Total stop_position nodes: {nodes_stop_position}\n")
    txtfile.write(f"Stop_position nodes with uic_ref: {stop_positions_with_uic_ref}\n\n")
    
    txtfile.write("MATCHING ANALYSIS\n")
    txtfile.write("----------------\n")
    txtfile.write(f"Platform nodes with matching stop_position (same uic_ref): {platforms_with_matching_stop}\n")
    txtfile.write(f"Platform nodes with uic_ref but no matching stop_position: {platforms_without_matching_stop}\n\n")
    
    # Write examples of platforms without matching stops
    txtfile.write("EXAMPLES OF PLATFORM NODES WITH NO MATCHING STOP_POSITION\n")
    txtfile.write("------------------------------------------------------\n")
    for i, example in enumerate(unmatched_platform_examples, 1):
        txtfile.write(f"Example {i}:\n")
        txtfile.write(f"  Node ID: {example['id']}\n")
        txtfile.write(f"  Location: {example['lat']}, {example['lon']}\n")
        txtfile.write(f"  UIC Reference: {example['uic_ref']}\n")
        txtfile.write(f"  Name: {example['name']}\n\n")

print(f"Text report has been started in {txt_filename}")

# New Route Analysis Section
print("Performing route analysis...")

# Read the CSV file with node-route information
try:
    # Read the CSV data into a pandas DataFrame
    routes_df = pd.read_csv("osm_nodes_with_routes.csv")
    
    # Basic route statistics
    total_routes = routes_df['route_name'].nunique()
    total_route_connections = len(routes_df)
    nodes_with_routes = routes_df['node_id'].nunique()
    
    # Analyze nodes by number of routes they serve
    node_route_counts = routes_df.groupby('node_id').size().reset_index(name='route_count')
    busiest_nodes = node_route_counts.nlargest(10, 'route_count')
    
    # Get additional data for the busiest nodes
    busiest_nodes_data = []
    for _, row in busiest_nodes.iterrows():
        node_id = row['node_id']
        # Get the first instance of this node to extract its type and uic_ref
        node_info = routes_df[routes_df['node_id'] == node_id].iloc[0]
        
        # Find the node in the XML file and get its name
        node_name = "N/A"
        node_element = root.find(f".//node[@id='{node_id}']")
        if node_element is not None:
            name_tag = node_element.find("tag[@k='name']")
            if name_tag is not None:
                node_name = name_tag.get('v')
        
        busiest_nodes_data.append({
            'node_id': node_id,
            'route_count': row['route_count'],
            'node_type': node_info['node_type'],
            'uic_ref': node_info['uic_ref'],
            'name': node_name
        })
    
    # Analyze route distribution by node type
    routes_by_node_type = routes_df.groupby('node_type')['route_name'].nunique().reset_index()
    
    # Analyze routes by direction (if available)
    direction_analysis = None
    if 'direction_id' in routes_df.columns and not routes_df['direction_id'].isna().all():
        direction_counts = routes_df['direction_id'].value_counts().to_dict()
        direction_analysis = {
            'direction_0': direction_counts.get(0, 0),
            'direction_1': direction_counts.get(1, 0),
            'direction_unknown': routes_df['direction_id'].isna().sum()
        }
    
    # Analyze route types (extract from route_name)
    route_types = defaultdict(int)
    for route_name in routes_df['route_name'].dropna().unique():
        route_name = str(route_name).lower()
        if 'bus' in route_name:
            route_types['bus'] += 1
        elif 'tram' in route_name:
            route_types['tram'] += 1
        elif 'train' in route_name or 'rail' in route_name:
            route_types['train'] += 1
        elif 'ferry' in route_name:
            route_types['ferry'] += 1
        else:
            route_types['other'] += 1
    
    # Find routes with the most stops
    routes_stop_counts = routes_df.groupby('route_name').size().reset_index(name='stop_count')
    routes_with_most_stops = routes_stop_counts.nlargest(10, 'stop_count')
    
    # Append to the existing text report file
    with open(txt_filename, "a", encoding="utf-8") as txtfile:
        txtfile.write("\n\nOSM ROUTE ANALYSIS REPORT\n")
        txtfile.write("=======================\n\n")
        
        # Basic statistics
        txtfile.write("BASIC ROUTE STATISTICS\n")
        txtfile.write("---------------------\n")
        txtfile.write(f"Total unique routes: {total_routes}\n")
        txtfile.write(f"Total node-route connections: {total_route_connections}\n")
        txtfile.write(f"Nodes that serve at least one route: {nodes_with_routes}\n")
        txtfile.write(f"Average routes per node: {total_route_connections / nodes_with_routes:.2f}\n\n")
        
        # Busiest nodes
        txtfile.write("TOP 10 BUSIEST TRANSIT NODES\n")
        txtfile.write("--------------------------\n")
        for i, node in enumerate(busiest_nodes_data, 1):
            txtfile.write(f"{i}. Node ID: {node['node_id']}\n")
            txtfile.write(f"   Name: {node['name']}\n")
            txtfile.write(f"   Routes served: {node['route_count']}\n")
            txtfile.write(f"   Node type: {node['node_type']}\n")
            txtfile.write(f"   UIC reference: {node['uic_ref'] if pd.notna(node['uic_ref']) else 'N/A'}\n\n")
        
        # Routes by node type
        txtfile.write("ROUTE DISTRIBUTION BY NODE TYPE\n")
        txtfile.write("-----------------------------\n")
        for _, row in routes_by_node_type.iterrows():
            txtfile.write(f"Node type '{row['node_type']}': {row['route_name']} unique routes\n")
        txtfile.write("\n")
        
        # Direction analysis
        if direction_analysis:
            txtfile.write("ROUTE DIRECTION ANALYSIS\n")
            txtfile.write("-----------------------\n")
            txtfile.write(f"Direction 0 (typically outbound): {direction_analysis['direction_0']} connections\n")
            txtfile.write(f"Direction 1 (typically inbound): {direction_analysis['direction_1']} connections\n")
            txtfile.write(f"Unknown direction: {direction_analysis['direction_unknown']} connections\n\n")
        
        # Route types
        txtfile.write("ROUTE TYPES BREAKDOWN\n")
        txtfile.write("--------------------\n")
        for route_type, count in route_types.items():
            txtfile.write(f"{route_type.capitalize()} routes: {count}\n")
        txtfile.write("\n")
        
        # Routes with most stops
        txtfile.write("TOP 10 ROUTES WITH MOST STOPS\n")
        txtfile.write("---------------------------\n")
        for i, row in enumerate(routes_with_most_stops.iterrows(), 1):
            index, data = row
            txtfile.write(f"{i}. {data['route_name']}: {data['stop_count']} stops\n")
    
    print(f"Route analysis appended to {txt_filename}")
    
    # Create a plot showing just the distribution of routes per stop
    plt.figure(figsize=(10, 6))
    
    # Count frequency of each route count
    route_count_distribution = node_route_counts['route_count'].value_counts().sort_index()
    plt.bar(route_count_distribution.index, route_count_distribution.values)
    plt.title('Distribution of Routes per Stop', fontsize=14)
    plt.xlabel('Number of Routes', fontsize=12)
    plt.ylabel('Number of Stops', fontsize=12)
    # For better readability if there are many route counts
    plt.xlim(left=0)
    plt.grid(axis='y', linestyle='--', alpha=0.7)
    
    # Create plots directory if it doesn't exist
    if not os.path.exists('plots'):
        os.makedirs('plots')
    
    # Save the visualization to the plots folder
    plt.tight_layout()
    plt.savefig('plots/routes_per_stop_distribution.png', dpi=300)
    print("Routes per stop distribution plot saved to plots/routes_per_stop_distribution.png")

except FileNotFoundError:
    print("Warning: osm_nodes_with_routes.csv not found. Route analysis skipped.")
except Exception as e:
    print(f"Error during route analysis: {e}")

print(f"All analysis reports have been consolidated into {txt_filename}")

