import pandas as pd
import logging
from collections import defaultdict
from tqdm import tqdm
from math import radians, sin, cos, sqrt, atan2
import xml.etree.ElementTree as ET
# Setup logging
logger = logging.getLogger(__name__)

def haversine_distance(lat1, lon1, lat2, lon2):
    """
    Calculate the Haversine distance (in meters) between two points.
    """
    try:
        R = 6371000.0  # Earth radius in meters
        lat1, lon1, lat2, lon2 = map(float, [lat1, lon1, lat2, lon2])
        lat1, lon1, lat2, lon2 = map(radians, [lat1, lon1, lat2, lon2])
        dlat = lat2 - lat1
        dlon = lon2 - lon1
        a = sin(dlat/2)**2 + cos(lat1)*cos(lat2)*sin(dlon/2)**2
        c = 2 * atan2(sqrt(a), sqrt(1 - a))
        return R * c
    except Exception as e:
        logger.error(f"Error in haversine_distance: {e}")
        return None

def route_matching(unmatched_df, xml_nodes, max_distance=50):
    """
    Two-stage route matching.
    
    Stage 1:
      Match entries if they have the same UIC reference, same route, and same direction.
    
    Stage 2:
      Match entries if they have the same route and direction, and are less than max_distance meters apart.
    
    Parameters:
        unmatched_df (pd.DataFrame): CSV rows with coordinate, route, and direction info.
        xml_nodes (dict): Dictionary where keys are (lat, lon) tuples and values are XML node dictionaries.
        max_distance (float): Maximum allowed distance (meters) for a match in Stage 2.

    Returns:
        List[dict]: List of match records (dictionaries) with candidate pool info.
    """
    matches = []
    
    # Load route information
    logger.info("Loading route information from CSV files...")
    try:
        # Load ATLAS routes data
        atlas_routes = pd.read_csv("atlas_routes.csv")
                
        # Create a mapping from sloid to route/direction
        atlas_route_map = defaultdict(list)
        for _, row in atlas_routes.iterrows():
            if pd.notna(row['sloid']):
                # Use route_id instead of route_short_name for matching
                atlas_route_map[row['sloid']].append({
                    'route': row['route_id'],  # Changed from route_short_name to route_id
                    'direction': str(row['direction_id'])
                })
        logger.info(f"Loaded {len(atlas_routes)} ATLAS route entries for {len(atlas_route_map)} stops")
        
        # Load OSM nodes with routes
        osm_routes = pd.read_csv("osm_nodes_with_routes.csv")
        
        
        # Create a mapping from node_id to route/direction
        osm_route_map = defaultdict(list)
        for _, row in osm_routes.iterrows():
            if pd.notna(row['node_id']) and pd.notna(row['direction_id']) and pd.notna(row['gtfs_route_id']):
                # Use only the gtfs_route_id for matching
                # Convert direction_id to integer string format to match ATLAS format
                direction = str(int(float(row['direction_id'])))
                osm_route_map[row['node_id']].append({
                    'route': row['gtfs_route_id'],
                    'direction': direction
                })
            
        logger.info(f"Loaded {len(osm_routes)} OSM route entries for {len(osm_route_map)} nodes")
    
    except Exception as e:
        logger.error(f"Error loading route data: {e}")
        logger.warning("Proceeding with route matching without additional route data")
    
    operator_mismatches = []
    # Track unique organization mismatches
    unique_org_mismatches = set()
    
    # Helper function to create match dictionary and handle operator mismatches
    def create_match_dict(csv_row, osm_lat, osm_lon, osm_node, distance, match_type, 
                          matching_notes, candidate_pool_size, route, direction, business_org_abbr):
        """
        Create a standardized match dictionary and handle operator mismatches.
        """
        osm_route = osm_node.get('tags', {}).get('route', '')
        osm_direction = osm_node.get('tags', {}).get('direction', '')
        osm_operator = osm_node.get('tags', {}).get('operator', '')
        
        # Check operator match if OSM operator is not null
        operator_match = True
        if osm_operator:
            operator_match = (osm_operator == business_org_abbr)
        
        # Include route and direction in the match_type
        match_type_with_route = f"{match_type}_{route}_{direction}"
        
        match_dict = {
            'sloid': csv_row['sloid'],
            'number': csv_row.get('number'),  # UIC reference
            'csv_lat': csv_row['wgs84North'],
            'csv_lon': csv_row['wgs84East'],
            'csv_route': route,
            'csv_direction': direction,
            'csv_business_org_abbr': business_org_abbr,
            'osm_lat': osm_lat,
            'osm_lon': osm_lon,
            'distance_m': distance,
            'osm_node_id': osm_node['node_id'],
            'osm_route': osm_route,
            'osm_direction': osm_direction,
            'osm_operator': osm_operator,
            'osm_uic_ref': osm_node['tags'].get('uic_ref', ''),
            'match_type': match_type_with_route,
            'candidate_pool_size': candidate_pool_size,
            'matching_notes': matching_notes
        }
        
        # If there's an operator mismatch, record it
        if not operator_match and osm_operator:
            match_dict['matching_notes'] += f" [OPERATOR MISMATCH: OSM={osm_operator}, ATLAS={business_org_abbr}]"
            operator_mismatches.append(match_dict.copy())
            # Add to unique mismatches set
            unique_org_mismatches.add((osm_operator, business_org_abbr))
        
        return match_dict
    
    # Group OSM nodes by UIC reference, route, and direction for Stage 1
    osm_by_uic_route_dir = defaultdict(list)
    # Group OSM nodes by route and direction for Stage 2
    osm_by_route_dir = defaultdict(list)
    
    logger.info("Building OSM node indexes for route matching...")
    # Debug: Print key types to diagnose the issue
    print(f"DEBUG - OSM route map key types: {type(next(iter(osm_route_map.keys()))) if osm_route_map else 'No keys'}")

    for (xml_lat, xml_lon), xml_node in xml_nodes.items():
        node_id = xml_node['node_id']
        uic_ref = xml_node['tags'].get('uic_ref', '')
        
        # Debug node_id and look for our specific case
        if uic_ref == '8511123':
            print(f"DEBUG - Found node with UIC 8511123: node_id={node_id}, type={type(node_id)}")
            print(f"DEBUG - Is in osm_route_map? {node_id in osm_route_map}")
            # Try converting to int if it's a string
            if isinstance(node_id, str) and node_id.isdigit():
                int_node_id = int(node_id)
                print(f"DEBUG - Is int version in osm_route_map? {int_node_id in osm_route_map}")
        
        # Convert node_id to int if it's a string (to match osm_route_map keys)
        if isinstance(node_id, str) and node_id.isdigit():
            node_id = int(node_id)
        
        # Get routes and directions from loaded OSM route data if available
        route_directions = []  # Initialize empty list by default
        if node_id in osm_route_map:
            route_directions = osm_route_map[node_id]
            
            # Debug matching for our specific case
            if uic_ref == '8511123':
                print(f"DEBUG - Found route_directions for node {node_id}: {route_directions}")
        
        # Add to appropriate index for each route/direction
        for rd in route_directions:
            route = rd['route']
            direction = rd['direction']
            
            if uic_ref:
                # Add this entry to the UIC index
                osm_by_uic_route_dir[(uic_ref, route, direction)].append({
                    'lat': xml_lat,
                    'lon': xml_lon,
                    'node': xml_node
                })
            
            # Add to route/direction index
            osm_by_route_dir[(route, direction)].append({
                'lat': xml_lat,
                'lon': xml_lon,
                'node': xml_node
            })

    # Keep track of matched ATLAS entries to exclude from later stages
    matched_sloids = set()
    
    # Combine Stage 1 and Stage 2 into a single loop for efficiency
    logger.info("Starting combined route matching...")
    for idx, csv_row in tqdm(unmatched_df.iterrows(), total=len(unmatched_df), desc="Route Matching"):
        try:
            csv_lat = float(csv_row['wgs84North'])
            csv_lon = float(csv_row['wgs84East'])
            sloid = csv_row['sloid']
            uic_ref = str(csv_row.get('number', '')).strip() if pd.notna(csv_row.get('number')) else ""
            business_org_abbr = str(csv_row.get('servicePointBusinessOrganisationAbbreviationEn', '')).strip() if pd.notna(csv_row.get('servicePointBusinessOrganisationAbbreviationEn')) else ""
            
            # Get routes and directions from atlas_route_map if available
            route_directions = []
            if 'atlas_route_map' in locals() and sloid in atlas_route_map:
                route_directions = atlas_route_map[sloid]
            
            # Skip if no route/direction info
            if not route_directions:
                continue
            
            matched = False
            
            # Try each route/direction combination
            for rd in route_directions:
                route = rd['route']
                direction = rd['direction']
                
                # STAGE 1: Try to match by UIC reference, route, and direction first
                if uic_ref:
                    key = (uic_ref, route, direction)
                    
                    # DEBUG: Special case logging for our problem pair
                    if uic_ref == '8511123' and route == '92-16-B-j24-1' and direction == '0':
                        print(f"DEBUG - Looking for Stage 1 match with key {key}")
                        print(f"DEBUG - Using osm_by_uic_route_dir index keys: {list(osm_by_uic_route_dir.keys())[:10]} ...")
                    
                    if key in osm_by_uic_route_dir:
                        candidates = osm_by_uic_route_dir[key]
                        
                        # Find the closest matching node
                        min_distance = float('inf')
                        best_match = None
                        
                        for candidate in candidates:
                            distance = haversine_distance(csv_lat, csv_lon, candidate['lat'], candidate['lon'])
                            if distance is not None and distance < min_distance:
                                min_distance = distance
                                best_match = candidate
                        
                        if best_match:
                            match_dict = create_match_dict(
                                csv_row, best_match['lat'], best_match['lon'], best_match['node'], min_distance,
                                'route_matching_1', "Exact UIC reference, route, and direction match",
                                len(candidates), route, direction, business_org_abbr
                            )
                            matches.append(match_dict)
                            matched_sloids.add(csv_row['sloid'])
                            matched = True
                            break  # Found a match with this route/direction, no need to try others
                
                # STAGE 2: If no Stage 1 match, try by route and direction within max_distance
                if not matched:
                    # Look for matching OSM nodes with the same route and direction
                    key = (route, direction)
                    if key in osm_by_route_dir:
                        candidates = []
                        
                        # Find all candidates within max_distance
                        for candidate in osm_by_route_dir[key]:
                            distance = haversine_distance(csv_lat, csv_lon, candidate['lat'], candidate['lon'])
                            if distance is not None and distance <= max_distance:
                                candidates.append({
                                    'lat': candidate['lat'],
                                    'lon': candidate['lon'],
                                    'node': candidate['node'],
                                    'distance': distance
                                })
                        
                        if candidates:
                            # Sort by distance
                            candidates.sort(key=lambda x: x['distance'])
                            best_match = candidates[0]
                            
                            match_dict = create_match_dict(
                                csv_row, best_match['lat'], best_match['lon'], best_match['node'], best_match['distance'],
                                'route_matching_2', f"Same route and direction match within {max_distance}m",
                                len(candidates), route, direction, business_org_abbr
                            )
                            matches.append(match_dict)
                            matched = True
                            break  # Found a match with this route/direction, no need to try others
                        
        except Exception as e:
            logger.error(f"Error in route matching for row {idx}: {e}")
            continue
    
    # Generate and print matching statistics
    logger.info(f"Route matching complete: {len(matches)} total matches found")

    # Separate matches by type
    stage1_matches = [m for m in matches if m['match_type'].startswith('route_matching_1_')]
    stage2_matches = [m for m in matches if m['match_type'].startswith('route_matching_2_')]

    # Print count statistics
    print("\n===== ROUTE MATCHING STATISTICS =====")
    print(f"Total ATLAS entries processed: {len(unmatched_df)}")
    print(f"Total matches found: {len(matches)} ({len(matches)/len(unmatched_df)*100:.2f}% of ATLAS entries)")
    print(f"  - Stage 1 (UIC+Route+Direction): {len(stage1_matches)} ({len(stage1_matches)/len(matches)*100:.2f}% of matches)")
    print(f"  - Stage 2 (Route+Direction+Distance): {len(stage2_matches)} ({len(stage2_matches)/len(matches)*100:.2f}% of matches)")

    # Print average distance stats
    distances = [m['distance_m'] for m in matches if m['distance_m'] is not None]
    if distances:
        avg_distance = sum(distances) / len(distances)
        max_distance = max(distances)
        print(f"\nDistance statistics:")
        print(f"  - Average match distance: {avg_distance:.2f} meters")
        print(f"  - Maximum match distance: {max_distance:.2f} meters")

    # Print specific information about our test case
    test_matches = [m for m in matches if m['sloid'] == 'ch:1:sloid:11123:0:413901']
    if test_matches:
        print("\nTest case match (sloid = ch:1:sloid:11123:0:413901):")
        for m in test_matches:
            print(f"  - Match type: {m['match_type']}")
            print(f"  - Distance: {m['distance_m']:.2f} meters")
            print(f"  - Notes: {m['matching_notes']}")
    else:
        print("\nTest case (sloid = ch:1:sloid:11123:0:413901) did NOT match")

    print("=====================================\n")

    # Write unique organization mismatches to file
    if unique_org_mismatches:
        with open('route_org_mismatches_review.txt', 'w') as f:
            f.write("===== UNIQUE OPERATOR MISMATCHES (ROUTE MATCHING) =====\n")
            f.write(f"Found {len(unique_org_mismatches)} unique operator mismatches:\n\n")
            for i, (osm_op, atlas_op) in enumerate(sorted(unique_org_mismatches), 1):
                f.write(f"{i}. OSM: {osm_op} | ATLAS: {atlas_op}\n")
        
        logger.info(f"Wrote {len(unique_org_mismatches)} unique operator mismatches to route_org_mismatches_review.txt")
    
    return matches

def parse_osm_xml(xml_file):
    """
    Parse the OSM XML file and build a comprehensive data structure that can be used for all matching strategies:
      - all_nodes: Complete representation of all OSM nodes with coordinates and tags
      - uic_ref_dict: { uic_ref : [ {node_id, lat, lon, local_ref, tags}, ... ] }
      - name_index: { name_value : [node_info, ...] } for keys: 'name', 'uic_name', 'gtfs:name'
    """
    tree = ET.parse(xml_file)
    root = tree.getroot()
    all_nodes = {}
    uic_ref_dict = defaultdict(list)
    name_index = defaultdict(list)
    ref_tag_count = 0
    
    for node in root.iter("node"):
        node_id = node.get("id")
        try:
            lat = float(node.get("lat"))
            lon = float(node.get("lon"))
        except (ValueError, TypeError):
            continue
            
        local_ref = None
        tags = {}
        
        for tag in node.findall("tag"):
            k = tag.get("k")
            v = tag.get("v")
            
            tags[k] = v
            if k == "local_ref":
                local_ref = v
            elif k == "ref":
                ref_tag_count += 1
                if not local_ref:
                    local_ref = v
        
        # Create node entry that will be used in all data structures
        node_entry = {
            'node_id': node_id,
            'lat': lat,
            'lon': lon,
            'local_ref': local_ref,
            'tags': tags
        }
        
        # Store in main nodes dictionary by coordinates
        all_nodes[(lat, lon)] = node_entry
        
        # Build UIC reference index
        if "uic_ref" in tags:
            uic_val = tags["uic_ref"]
            uic_ref_dict[uic_val].append(node_entry)
        
        # Build name index
        for key in ['name', 'uic_name', 'gtfs:name']:
            if key in tags:
                name_index[tags[key]].append(node_entry)
    
    logger.info(f"Parsed OSM XML: {len(all_nodes)} nodes, {len(uic_ref_dict)} uic_ref entries; {ref_tag_count} 'ref' tags encountered.")
    return all_nodes, uic_ref_dict, name_index
"""
def main():
    atlas_csv_file = "stops_ATLAS.csv"
    osm_xml_file = "osm_data.xml"
    
    atlas_df = pd.read_csv(atlas_csv_file, sep=";")
    all_osm_nodes, uic_ref_dict, name_index = parse_osm_xml(osm_xml_file)
    route_matches = route_matching(atlas_df, all_osm_nodes, max_distance=50)
    
    # Save route matches to CSV
    if route_matches:
        matches_df = pd.DataFrame(route_matches)
        matches_df.to_csv("route_matches.csv", index=False)
        logger.info(f"Saved {len(route_matches)} route matches to route_matches.csv")
    else:
        logger.warning("No route matches found to save")

# Add this at the end of the file (outside any function)
if __name__ == "__main__":
    main()
"""