// Global variables
var map;
var osmLayer;
var markersLayer = L.layerGroup();
var linesLayer = L.layerGroup();
var topNLayer = L.layerGroup(); // For top N distances overlay
var stopsById = {};   // Global store for stops by id.
// manual matching variables are now defined in manual-matching.js

// Function to format route data for display
function formatRouteInfo(routes) {
    // Handle empty or undefined routes
    if (!routes) {
        return "<i>No route information available</i>";
    }
    
    // Parse JSON string if needed
    if (typeof routes === 'string') {
        try {
            routes = JSON.parse(routes);
        } catch (e) {
            return "<i>Error parsing route data</i>";
        }
    }
    
    // Handle empty arrays
    if (Array.isArray(routes) && routes.length === 0) {
        return "<i>No route information available</i>";
    }
    
    try {
        // Group routes by route_id
        const routeGroups = {};
        
        if (Array.isArray(routes)) {
            routes.filter(route => route).forEach(route => {
                const routeId = route.route_id || "unknown";
                const routeName = route.route_short_name || route.route_name || route.route_id || "Unnamed Route";
                
                if (!routeGroups[routeId]) {
                    routeGroups[routeId] = {
                        name: routeName,
                        directions: [],
                        routeId: routeId
                    };
                }
                
                // Only add direction if it's defined and not already in the list
                if (route.direction_id !== undefined && !routeGroups[routeId].directions.includes(route.direction_id)) {
                    routeGroups[routeId].directions.push(route.direction_id);
                }
            });
        }
        
        // Generate HTML list items from grouped routes
        const routeItems = Object.values(routeGroups).map(group => {
            const routeName = group.name;
            const directions = group.directions.sort();
            const directionsStr = directions.length > 0 ? `Dir:${directions.join(',')}` : "";
            
            // Make route ID clickable to filter by it
            const routeIdLink = group.routeId !== "unknown" 
                ? `(ID: <a href="#" onclick="filterByRoute('${group.routeId}', '${directions.join(',')}'); return false;">${group.routeId}</a>)` 
                : "";
            
            return `<li>${routeName} ${routeIdLink} ${directionsStr}</li>`;
        }).join('');
        
        return routeItems.length > 0 
            ? `<ul style='margin-top: 5px; padding-left: 15px;'>${routeItems}</ul>` 
            : "<i>No route information available</i>";
    } catch (e) {
        return "<i>Error processing route data</i>";
    }
}

// Function to filter by route
function filterByRoute(routeId, directions) {
    if (!routeId) {
        alert("No route ID available.");
        return;
    }
    
    // Check if this route filter is already applied
    var isDuplicate = false;
    for (var i = 0; i < activeFilters.station.length; i++) {
        if (activeFilters.stationTypes[i] === 'route' && 
            activeFilters.station[i] === routeId &&
            activeFilters.routeDirections[i] === directions) {
            isDuplicate = true;
            break;
        }
    }
    
    if (isDuplicate) {
        alert("This route filter is already applied: Route: " + routeId + (directions ? ", Direction: " + directions : ""));
        return;
    }
    
    // Set filter type to 'route' for the UI
    $('#filterTypeSelector').val('route');
    toggleFilterInputs();
    
    // Add the route filter
    activeFilters.station.push(routeId);
    activeFilters.stationTypes.push('route');
    activeFilters.routeDirections.push(directions || '');
    activeFilters.filterType = 'route';
    
    updateFiltersUI();
    loadDataForViewport();
}

// New function to format routes with matched vs. unmatched categorization
function formatRoutesDisplay(atlasRoutes, osmRoutes, isOsmNode = false) {
    // Normalize routes to arrays and handle edge cases
    const atlasRoutesArray = Array.isArray(atlasRoutes) ? atlasRoutes : [];
    const osmRoutesArray = Array.isArray(osmRoutes) ? osmRoutes : [];
    
    // If both are empty, return a simple message without a collapsible
    if (atlasRoutesArray.length === 0 && osmRoutesArray.length === 0) {
        return "<i>No route information available</i>";
    }
    
    // Find matched routes by comparing route_id and direction_id
    const matchedRoutes = [];
    const atlasOnlyRoutes = [...atlasRoutesArray];
    const osmOnlyRoutes = [...osmRoutesArray];
    
    // Find routes that match between ATLAS and OSM
    atlasRoutesArray.forEach(atlasRoute => {
        if (!atlasRoute || !atlasRoute.route_id) return;
        
        const matchingOsmRouteIndex = osmRoutesArray.findIndex(osmRoute => 
            osmRoute && 
            osmRoute.route_id === atlasRoute.route_id && 
            osmRoute.direction_id === atlasRoute.direction_id
        );
        
        if (matchingOsmRouteIndex !== -1) {
            // Found a match
            const osmRoute = osmRoutesArray[matchingOsmRouteIndex];
            matchedRoutes.push({
                route_id: atlasRoute.route_id,
                direction_id: atlasRoute.direction_id,
                route_short_name: atlasRoute.route_short_name || osmRoute.route_name,
                route_long_name: atlasRoute.route_long_name,
                route_name: osmRoute.route_name
            });
            
            // Remove from the "only" arrays
            const atlasIndex = atlasOnlyRoutes.findIndex(r => 
                r && r.route_id === atlasRoute.route_id && r.direction_id === atlasRoute.direction_id
            );
            if (atlasIndex !== -1) atlasOnlyRoutes.splice(atlasIndex, 1);
            
            const osmIndex = osmOnlyRoutes.findIndex(r => 
                r && r.route_id === osmRoute.route_id && r.direction_id === osmRoute.direction_id
            );
            if (osmIndex !== -1) osmOnlyRoutes.splice(osmIndex, 1);
        }
    });
    
    // Build the HTML for the different sections
    let html = '';
    
    // For OSM nodes, only display matched and OSM-only routes
    if (isOsmNode) {
        // If no routes to display for OSM node, return simple message
        if (matchedRoutes.length === 0 && osmOnlyRoutes.length === 0) {
            return "<i>No route information available</i>";
        }
        
        // Matched routes section
        if (matchedRoutes.length > 0) {
            html += `<div><strong>Matched Routes:</strong>${formatRouteInfo(matchedRoutes)}</div>`;
        }
        
        // OSM-only routes section
        if (osmOnlyRoutes.length > 0) {
            html += `<div><strong>OSM-only Routes:</strong>${formatRouteInfo(osmOnlyRoutes)}</div>`;
        }
        
        return html;
    } 
    // For ATLAS nodes, display all route types
    else {
        // Matched routes section
        if (matchedRoutes.length > 0) {
            html += `<div><strong>Matched Routes:</strong>${formatRouteInfo(matchedRoutes)}</div>`;
        }
        
        // ATLAS-only routes section
        if (atlasOnlyRoutes.length > 0) {
            html += `<div><strong>ATLAS-only Routes:</strong>${formatRouteInfo(atlasOnlyRoutes)}</div>`;
        }
        
        // OSM-only routes section (for matched ATLAS stops, showing all routes)
        if (osmOnlyRoutes.length > 0) {
            html += `<div><strong>OSM-only Routes:</strong>${formatRouteInfo(osmOnlyRoutes)}</div>`;
        }
        
        return html;
    }
}

// Function to create a collapsible section
function createCollapsible(title, content, isExpanded = false) {
    // If content is simply the "No route information available" message, return it directly with a line break
    if (content === "<i>No route information available</i>") {
        return content + "<br>";
    }
    
    const id = 'collapse-' + Math.random().toString(36).substring(2, 9); // Generate random ID
    return `
        <div class="collapsible">
            <button type="button" class="btn btn-sm btn-outline-secondary" 
                    onclick="toggleCollapsible('${id}')">
                ${title} <span id="${id}-arrow">${isExpanded ? '▲' : '▼'}</span>
            </button>
            <div id="${id}" class="collapsible-content" 
                style="display: ${isExpanded ? 'block' : 'none'}; 
                      margin-top: 5px; 
                      border-left: 3px solid #ccc; 
                      padding-left: 8px;
                      max-height: 150px;
                      overflow-y: auto;
                      scrollbar-width: thin;
                      scrollbar-color: #6c757d #e9ecef;">
                ${content}
            </div>
        </div>
    `;
}

// Helper function to toggle collapsible sections
function toggleCollapsible(id) {
    const element = document.getElementById(id);
    const arrow = document.getElementById(id + '-arrow');
    if (element.style.display === 'none') {
        element.style.display = 'block';
        arrow.textContent = '▲';
    } else {
        element.style.display = 'none';
        arrow.textContent = '▼';
    }
}

// Create clickable link for sloid or node_id filtering
function createFilterLink(value, type, displayText = null) {
    if (!value) return "N/A";
    const text = displayText || value;
    return `<a href="#" onclick="addCustomFilter('${value}', '${type}'); return false;">${text}</a>`;
}

// Add a custom filter programmatically
function addCustomFilter(value, filterType) {
    if (!value) return;
    
    // Check if this filter is already applied
    var isDuplicate = false;
    for (var i = 0; i < activeFilters.station.length; i++) {
        if (activeFilters.stationTypes[i] === filterType && 
            activeFilters.station[i] === value) {
            isDuplicate = true;
            break;
        }
    }
    
    if (isDuplicate) {
        alert("This filter is already applied: " + filterType + ": " + value);
        return;
    }
    
    // Add the filter
    activeFilters.station.push(value);
    activeFilters.stationTypes.push(filterType);
    activeFilters.routeDirections.push('');
    
    updateFiltersUI();
    loadDataForViewport();
}

// Function to create unified popup content for matched stops
function createUnifiedPopupContent(stop) {
    let popupHtml = '<div class="unified-popup">';
    // Unified container of bubbles
    popupHtml += '<div class="matches-container">';

    // Case: OSM-centric with multiple ATLAS matches
    if (stop.stop_type === 'matched' && stop.is_osm_node && Array.isArray(stop.atlas_matches)) {
        // OSM bubble
        popupHtml += `
            <div class="osm-match">
                <h5>OSM Node</h5>
                <table class="popup-table">
                    <tr><td>Node ID:</td><td>${createFilterLink(stop.osm_node_id, 'osm')}</td></tr>
                    <tr><td>Name:</td><td>${stop.osm_name || 'N/A'}</td></tr>
                    <tr><td>Network:</td><td>${stop.osm_network || 'N/A'}</td></tr>
                    <tr><td>Coordinates:</td><td>(${stop.osm_lat || stop.lat}, ${stop.osm_lon || stop.lon})</td></tr>
                </table>
                <div class="route-section">${createCollapsible("Routes", formatRouteInfo(stop.routes_osm))}</div>
            </div>`;
        // ATLAS bubbles
        stop.atlas_matches.forEach((atlasMatch, index) => {
            const matchNumber = stop.atlas_matches.length > 1 ? ` #${index + 1}` : '';
            popupHtml += `
                <div class="atlas-match">
                    <h5>ATLAS Stop${matchNumber}</h5>
                    <table class="popup-table">
                        <tr><td>Sloid:</td><td>${createFilterLink(atlasMatch.sloid, 'atlas')}</td></tr>
                        <tr><td>UIC Ref:</td><td>${createFilterLink(atlasMatch.uic_ref, 'station')}</td></tr>
                        <tr><td>Name:</td><td>${atlasMatch.atlas_designation || 'N/A'}</td></tr>
                        <tr><td>Distance:</td><td>${atlasMatch.distance_m ? parseFloat(atlasMatch.distance_m).toFixed(1) + ' m' : 'N/A'}</td></tr>
                    </table>
                    <div class="route-section">${createCollapsible("Routes", formatRouteInfo(atlasMatch.routes_atlas))}</div>
                </div>`;
        });
    }
    // Case: ATLAS-centric with multiple OSM matches
    else if (stop.stop_type === 'matched' && Array.isArray(stop.osm_matches)) {
        // ATLAS bubble
        popupHtml += `
            <div class="atlas-match">
                <h5>ATLAS Stop</h5>
                <table class="popup-table">
                    <tr><td>Sloid:</td><td>${createFilterLink(stop.sloid, 'atlas')}</td></tr>
                    <tr><td>UIC Ref:</td><td>${createFilterLink(stop.uic_ref, 'station')}</td></tr>
                    <tr><td>Name:</td><td>${stop.atlas_designation || 'N/A'}</td></tr>
                    <tr><td>Coordinates:</td><td>(${stop.atlas_lat}, ${stop.atlas_lon})</td></tr>
                </table>
                <div class="route-section">${createCollapsible("Routes", formatRouteInfo(stop.routes_atlas))}</div>
            </div>`;
        // OSM bubbles
        stop.osm_matches.forEach((osmMatch, index) => {
            const matchNumber = stop.osm_matches.length > 1 ? ` #${index + 1}` : '';
            popupHtml += `
                <div class="osm-match">
                    <h5>OSM Node${matchNumber}</h5>
                    <table class="popup-table">
                        <tr><td>Node ID:</td><td>${createFilterLink(osmMatch.osm_node_id, 'osm')}</td></tr>
                        <tr><td>Name:</td><td>${osmMatch.osm_name || 'N/A'}</td></tr>
                        <tr><td>Network:</td><td>${osmMatch.osm_network || 'N/A'}</td></tr>
                        <tr><td>Distance:</td><td>${osmMatch.distance_m ? parseFloat(osmMatch.distance_m).toFixed(1) + ' m' : 'N/A'}</td></tr>
                    </table>
                    <div class="route-section">${createCollapsible("Routes", formatRouteInfo(osmMatch.routes_osm))}</div>
                </div>`;
        });
    }
    // Case: simple one-to-one matched stop
    else if (stop.stop_type === 'matched' && stop.atlas_lat && stop.osm_lat) {
        // ATLAS bubble
        popupHtml += `
            <div class="atlas-match">
                <h5>ATLAS Stop</h5>
                <table class="popup-table">
                    <tr><td>Sloid:</td><td>${createFilterLink(stop.sloid, 'atlas')}</td></tr>
                    <tr><td>UIC Ref:</td><td>${createFilterLink(stop.uic_ref, 'station')}</td></tr>
                    <tr><td>Name:</td><td>${stop.atlas_designation || 'N/A'}</td></tr>
                    <tr><td>Coordinates:</td><td>(${stop.atlas_lat}, ${stop.atlas_lon})</td></tr>
                </table>
                <div class="route-section">${createCollapsible("Routes", formatRouteInfo(stop.routes_atlas))}</div>
            </div>`;
        // OSM bubble
        popupHtml += `
            <div class="osm-match">
                <h5>OSM Node</h5>
                <table class="popup-table">
                    <tr><td>Node ID:</td><td>${createFilterLink(stop.osm_node_id, 'osm')}</td></tr>
                    <tr><td>Name:</td><td>${stop.osm_name || 'N/A'}</td></tr>
                    <tr><td>Network:</td><td>${stop.osm_network || 'N/A'}</td></tr>
                    <tr><td>Distance:</td><td>${stop.distance_m ? parseFloat(stop.distance_m).toFixed(1) + ' m' : 'N/A'}</td></tr>
                </table>
                <div class="route-section">${createCollapsible("Routes", formatRouteInfo(stop.routes_osm))}</div>
            </div>`;
    }

    popupHtml += '</div>'; // Close matches-container
    popupHtml += '</div>'; // Close unified-popup
    return popupHtml;
}

// Unified global filters object
var activeFilters = {
    station: [],      // Combined station/stop filter 
    stationTypes: [], // Types corresponding to each station filter (station, atlas, osm)
    routeDirections: [], // Directions for route filters (empty string if no direction specified)
    filterType: 'station', // Type of filter: 'station', 'atlas', 'osm', or 'route'
    stopType: [],     // Matching filter: "matched", "unmatched"
    nodeType: [],     // Node type: "atlas", "osm"
    matchMethods: [], // Array for matching methods (exact, name, distance, manual)
    topN: null        // Top N distances value (if applicable)
};

// Helper function to extract and normalize the station identifier
function getStationIdentifier(stop, stopCategory) {
    // Use the unified uic_ref field if available
    if (stop.uic_ref) {
      return String(stop.uic_ref).trim().toLowerCase();
    }
    // For OSM stops that might not have a uic_ref, fallback to local_ref or node_id
    if (stopCategory === "osm") {
      if (stop.osm_local_ref) {
        return String(stop.osm_local_ref).trim().toLowerCase();
      } else if (stop.osm_node_id) {
        return String(stop.osm_node_id).trim().toLowerCase();
      }
    }
    return null;
}

function createOsmMarker(lat, lon, hasPublicTransportPlatform, color) {
    if (hasPublicTransportPlatform) {
        return L.marker([lat, lon], {
            icon: L.divIcon({
                html: `<svg width="12" height="12" viewBox="0 0 12 12">
                        <circle cx="6" cy="6" r="6" fill="${color}" fill-opacity="0.5" stroke="${color}" stroke-width="2"/>
                        <text x="6" y="8.5" text-anchor="middle" fill="white" font-size="8" font-weight="bold">P</text>
                      </svg>`,
                className: 'custom-div-icon',
                iconSize: [12, 12],
                iconAnchor: [6, 6]
            })
        });
    } else {
        return L.circleMarker([lat, lon], { 
            color: color, 
            radius: 6,
            fillOpacity: 0.5,
            weight: 2
        });
    }
}

// Function to initialize the map with event listeners that preserve popups during movement
function initMap() {
    map = L.map('map', {
        closePopupOnClick: false // Prevent map click from closing popups
    }).setView([47.3769, 8.5417], 13);
    
    osmLayer = L.tileLayer('https://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png', {
         maxZoom: 19,
         attribution: '© OpenStreetMap'
    });
    osmLayer.addTo(map);
    markersLayer.addTo(map);
    linesLayer.addTo(map);
    topNLayer.addTo(map);
    
    // Keep track of open popups
    var openPopups = [];
    map.on('popupopen', function(e) {
        openPopups.push(e.popup);
    });
    map.on('popupclose', function(e) {
        // Remove from the list and clean up lines
        openPopups = openPopups.filter(p => p !== e.popup);
        if (e.popup instanceof L.DraggablePopup && e.popup._line) {
            try { e.popup._removeLine(); } catch {};
        }
    });
    
    // Update popups and lines continuously during movement
    map.on('move', function() {
        if (window.updateAllPopupLines) window.updateAllPopupLines();
        openPopups.forEach(popup => {
            if (popup._updatePosition) popup._updatePosition();
        });
    });
    map.on('zoom', function() {
        if (window.updateAllPopupLines) window.updateAllPopupLines();
        openPopups.forEach(popup => {
            if (popup._updatePosition) popup._updatePosition();
        });
    });
    
    // After pan/zoom ends: reload data only if no popups are open
    map.on('moveend zoomend', function() {
        if (openPopups.length === 0) {
            loadDataForViewport();
        }
    });
}

// Update active filters UI (badges)
function updateFiltersUI() {
    var container = $('#activeFilters');
    container.empty();
    // Display stop type filters
    activeFilters.stopType.forEach(function(filter) {
        var badge = $('<span class="badge badge-primary mr-1 mb-1">' + filter +
            ' <a href="#" class="text-white remove-filter" data-type="stopType" data-filter="' + filter + '">x</a></span>');
        container.append(badge);
    });
    // Display node type filters
    activeFilters.nodeType.forEach(function(filter) {
        var badge = $('<span class="badge badge-success mr-1 mb-1">Node: ' + filter +
            ' <a href="#" class="text-white remove-filter" data-type="nodeType" data-filter="' + filter + '">x</a></span>');
        container.append(badge);
    });
    // Display matching method filters
    activeFilters.matchMethods.forEach(function(filter) {
        var displayName = filter;
        // Give more descriptive names to distance matching methods
        if (filter === 'distance_matching_1') displayName = 'Distance: Group Proximity';
        if (filter === 'distance_matching_2') displayName = 'Distance: Local Ref Match';
        if (filter === 'distance_matching_3a') displayName = 'Distance: Single Candidate';
        if (filter === 'distance_matching_3b') displayName = 'Distance: Relative Distance';
        if (filter === 'no_osm_within_100m') displayName = 'Distance: No OSM within 100m';
        
        // Give more descriptive names to route matching methods
        if (filter.startsWith('route_matching_1_')) displayName = 'Route: Same UIC, Route, Direction';
        if (filter.startsWith('route_matching_2_')) displayName = 'Route: Same Route, Direction within 50m';
        
        var badge = $('<span class="badge badge-secondary mr-1 mb-1">' + displayName +
            ' <a href="#" class="text-white remove-filter" data-type="matchMethod" data-filter="' + filter + '">x</a></span>');
        container.append(badge);
    });
    // Display station/node filters with proper type label
    activeFilters.station.forEach(function(filter, index) {
        // Set the proper label based on filter type - use the stored type for this filter
        var filterType = activeFilters.stationTypes[index] || activeFilters.filterType;
        var labelText = '';
        var badgeClass = '';
        
        switch(filterType) {
            case 'atlas':
                labelText = 'ATLAS SloidID: ';
                badgeClass = 'badge-warning';
                break;
            case 'osm':
                labelText = 'OSM Node ID: ';
                badgeClass = 'badge-info';
                break;
            case 'route':
                var direction = activeFilters.routeDirections[index] || '';
                labelText = 'Route ID: ';
                if (direction) {
                    labelText += filter + ', Direction: ' + direction;
                } else {
                    labelText += filter;
                }
                badgeClass = 'badge-danger';
                // Skip adding the filter value since we've already included it in the label
                var badge = $('<span class="badge ' + badgeClass + ' mr-1 mb-1">' + labelText +
                    ' <a href="#" class="text-white remove-filter" data-type="station" data-index="' + index + '">x</a></span>');
                container.append(badge);
                return; // Skip the rest of this iteration
                break;
            case 'station':
            default:
                labelText = 'UIC Number: ';
                badgeClass = 'badge-warning';
                break;
        }
        
        var badge = $('<span class="badge ' + badgeClass + ' mr-1 mb-1">' + labelText + filter +
            ' <a href="#" class="text-white remove-filter" data-type="station" data-index="' + index + '">x</a></span>');
        container.append(badge);
    });
    if(activeFilters.topN) {
        var badge = $('<span class="badge badge-info mr-1 mb-1">Top N Distances (' + activeFilters.topN + ') <a href="#" class="text-white remove-filter" data-type="topN" data-filter="topN">x</a></span>');
        container.append(badge);
    }
}

// Function to toggle between standard filter input and route filter input based on filter type
function toggleFilterInputs() {
    var filterType = $('#filterTypeSelector').val();
    
    if (filterType === 'route') {
        $('#standardFilterInput').hide();
        $('#routeFilterInput').show();
    } else {
        $('#standardFilterInput').show();
        $('#routeFilterInput').hide();
    }
    
    // Update placeholder text for standard filter
    var placeholder = "";
    switch(filterType) {
        case 'atlas':
            placeholder = "Enter ATLAS SloidID";
            break;
        case 'osm':
            placeholder = "Enter OSM Node ID";
            break;
        case 'route':
            placeholder = "Enter Route ID";
            break;
        case 'station':
        default:
            placeholder = "Enter UIC Number";
            break;
    }
    $('#stationFilter').attr('placeholder', placeholder);
}

// Function to add a new standard filter from the text input
function addTextFilter() {
    var filterVal = $('#stationFilter').val().trim();
    var filterType = $('#filterTypeSelector').val(); // Store the current filter type
    
    if (filterVal && !activeFilters.station.includes(filterVal)) {
        activeFilters.station.push(filterVal);
        activeFilters.stationTypes.push(filterType); // Store the type for this filter
        activeFilters.routeDirections.push(''); // No direction for standard filters
        $('#stationFilter').val(''); // Clear input field after adding
        updateFiltersUI();
        loadDataForViewport();
    }
}

// Function to add a new route+direction filter
function addRouteFilter() {
    var routeId = $('#routeIdFilter').val().trim();
    var direction = $('#directionFilter').val().trim();
    
    if (routeId) {
        // Check if this exact route+direction combination already exists
        var isDuplicate = false;
        for (var i = 0; i < activeFilters.station.length; i++) {
            if (activeFilters.stationTypes[i] === 'route' && 
                activeFilters.station[i] === routeId && 
                activeFilters.routeDirections[i] === direction) {
                isDuplicate = true;
                break;
            }
        }
        
        if (!isDuplicate) {
            activeFilters.station.push(routeId);
            activeFilters.stationTypes.push('route');
            activeFilters.routeDirections.push(direction);
            
            // Clear input fields after adding
            $('#routeIdFilter').val('');
            $('#directionFilter').val('');
            
            updateFiltersUI();
            loadDataForViewport();
        } else {
            alert('This route and direction combination is already filtered.');
        }
    } else {
        alert('Please enter a Route ID.');
    }
}

// Remove filter badge and update filters
$(document).on('click', '.remove-filter', function(e) {
    e.preventDefault();
    e.stopPropagation(); // Prevent event bubbling
    var type = $(this).data('type');
    var filter = $(this).data('filter');
    var index = $(this).data('index');

    console.log("Remove filter clicked:", type, filter, index);

    // Handle topN filter separately because it's not an array.
    if (type === "topN") {
        activeFilters.topN = null;
        $('#toggleTopNBtn').text("Activate Top N Filter");
        topNLayer.clearLayers();
        $('#topNDistancesMessage').empty();
        loadDataForViewport();
        updateFiltersUI();
        return false;
    }

    // For station filters, use the index to remove correctly
    if (type === "station" && index !== undefined) {
        if (index >= 0 && index < activeFilters.station.length) {
            activeFilters.station.splice(index, 1);
            activeFilters.stationTypes.splice(index, 1);
            activeFilters.routeDirections.splice(index, 1);
        }
    }
    // For other filters that are stored as arrays
    else if (activeFilters[type] && Array.isArray(activeFilters[type])) {
        activeFilters[type] = activeFilters[type].filter(function(f) { return f !== filter; });
    }
    
    if (type === "stopType") {
        $('.filter-stop-type[value="'+filter+'"]').prop('checked', false);
    } else if (type === "nodeType") {
        $('.filter-node-type[value="'+filter+'"]').prop('checked', false);
    } else if (type === "matchMethod") {
        // Uncheck both standard and distance matching checkboxes
        $('.filter-match-method[value="'+filter+'"]').prop('checked', false);
        $('.filter-distance-method[value="'+filter+'"]').prop('checked', false);
    }
    
    updateFiltersUI();
    loadDataForViewport();
});

function filterByStation(stopId, stopCategory) {
    var selectedStop = stopsById[stopId];
    if (!selectedStop) {
      alert("Stop data not found.");
      return;
    }
    
    // Always filter by UIC ref (station) when the "Filter by station" button is clicked
    var filterType = 'station';
    var filterValue = selectedStop.uic_ref || '';
    
    if (!filterValue) {
      alert("No UIC reference available for this stop.");
      return;
    }
    
    // Check if this filter is already applied
    var isDuplicate = false;
    for (var i = 0; i < activeFilters.station.length; i++) {
        if (activeFilters.stationTypes[i] === filterType && 
            activeFilters.station[i] === filterValue) {
            isDuplicate = true;
            break;
        }
    }
    
    if (isDuplicate) {
      alert("This filter is already applied: " + filterType + ": " + filterValue);
      return;
    }
    
    // Update the filter type dropdown for the UI (doesn't affect existing filters)
    $('#filterTypeSelector').val(filterType);
    toggleFilterInputs();
    
    // Add the filter to the existing filters rather than replacing them
    activeFilters.station.push(filterValue);
    activeFilters.stationTypes.push(filterType);
    activeFilters.routeDirections.push('');
    activeFilters.filterType = filterType;
    
    updateFiltersUI();
    loadDataForViewport();
}

// Update the activeFilters object from the UI inputs
function updateActiveFilters() {
    // Gather matching filters (stop type)
    activeFilters.stopType = $('.filter-stop-type:checked').map(function() {
        return this.value;
    }).get();
    
    // Gather node type filters for matched stops
    activeFilters.nodeType = $('.filter-node-type:checked').map(function() {
        return this.value;
    }).get();
    
    // Combined Station or Stop filter: read text input (if any)
    var filterVal = $('#stationFilter').val().trim();
    // Don't replace, just clear if empty
    if (filterVal === '') {
        $('#stationFilter').val('');
    }
    
    // Get the filter type from the dropdown
    activeFilters.filterType = $('#filterTypeSelector').val();
    
    // If 'matched' is selected, show matching methods and Top N distances (but do not update topN here)
    if(activeFilters.stopType.includes("matched")) {
        $('#matchingMethodContainer').show();
        $('#distanceMatchingContainer').show();  // Show distance matching specific filters
        $('#routeMatchingContainer').show();     // Show route matching specific filters
        $('#topNDistancesContainer').show();
        
        // Get standard match methods
        var standardMatchMethods = $('.filter-match-method:checked').map(function() {
            return this.value;
        }).get();
        
        // Get distance matching specific methods
        var distanceMatchMethods = $('.filter-distance-method:checked').map(function() {
            return this.value;
        }).get();
        
        // Get route matching specific methods
        var routeMatchMethods = $('.filter-route-method:checked').map(function() {
            return this.value;
        }).get();
        
        // Combine all matching methods
        activeFilters.matchMethods = standardMatchMethods.concat(distanceMatchMethods).concat(routeMatchMethods);
    } else {
        $('#matchingMethodContainer').hide();
        $('#distanceMatchingContainer').hide();  // Hide distance matching specific filters
        $('#routeMatchingContainer').hide();     // Hide route matching specific filters
        $('#topNDistancesContainer').hide();
        activeFilters.matchMethods = [];
        
        // If topN is active but matched is no longer selected, clear topN
        if(activeFilters.topN) {
            activeFilters.topN = null;
            $('#toggleTopNBtn').text("Activate Top N Filter");
            topNLayer.clearLayers();
            $('#topNDistancesMessage').empty();
        }
    }
    
    updateFiltersUI();
    loadDataForViewport();
    
    if(activeFilters.topN) {
        loadTopNMatches();
    } else {
        topNLayer.clearLayers();
        $('#topNDistancesMessage').empty();
    }
}

function loadTopNMatches() {
    topNLayer.clearLayers();
    $('#topNDistancesMessage').empty();
    
    if(activeFilters.topN && activeFilters.stopType.includes("matched")) {
        var params = { limit: activeFilters.topN };
        if(activeFilters.matchMethods.length > 0) {
            params.match_method = activeFilters.matchMethods.join(',');
        }
        
        // Add station filter values and type if available
        if(activeFilters.station.length > 0) {
            params.station_filter = activeFilters.station.join(',');
            params.filter_types = activeFilters.stationTypes.join(',');
            params.route_directions = activeFilters.routeDirections.join(',');
        }
        
        $.getJSON("/api/top_matches", params, function(data) {
            if(data.length === 0) {
                $('#topNDistancesMessage').html("<div class='alert alert-warning mt-2'>No matched nodes satisfy these conditions.</div>");
            } else {
                // Check if node type filtering is active
                var showAtlasNodes = activeFilters.nodeType.length === 0 || activeFilters.nodeType.indexOf("atlas") !== -1;
                var showOSMNodes = activeFilters.nodeType.length === 0 || activeFilters.nodeType.indexOf("osm") !== -1;
                
                data.forEach(function(stop) {
                    if(stop.stop_type === 'matched' && stop.atlas_lat && stop.atlas_lon && stop.osm_lat && stop.osm_lon) {
                        // Create a unified popup for this match
                        let unifiedPopup = createPopupWithOptions(createUnifiedPopupContent(stop));
                        
                        if(showAtlasNodes) {
                            var atlasMarker = L.circleMarker([parseFloat(stop.atlas_lat), parseFloat(stop.atlas_lon)], 
                                { color: "green", radius: 6 })
                                .bindPopup(unifiedPopup);
                            topNLayer.addLayer(atlasMarker);
                        }
                        if(showOSMNodes) {
                            var isPublicTransportPlatform = stop.osm_public_transport === 'platform';
                            var osmMarker = createOsmMarker(
                                parseFloat(stop.osm_lat), 
                                parseFloat(stop.osm_lon), 
                                isPublicTransportPlatform,
                                'blue'
                            );
                            osmMarker.bindPopup(unifiedPopup);
                            topNLayer.addLayer(osmMarker);
                        }
                        
                        // Only add connecting line if both node types are visible
                        if(showAtlasNodes && showOSMNodes) {
                            var line = L.polyline([
                                [parseFloat(stop.atlas_lat), parseFloat(stop.atlas_lon)],
                                [parseFloat(stop.osm_lat), parseFloat(stop.osm_lon)]
                            ], { color: "purple", dashArray: "5,5" });
                            topNLayer.addLayer(line);
                        }
                    }
                });
            }
        });
    }
}

// Helper function to customize popup options for all bindPopup calls
function createPopupWithOptions(content) {
    return L.draggablePopup({
        autoClose: false,
        closeOnClick: false,
        autoPan: false,
        maxWidth: 300,
        closeButton: true,
        className: 'customPopup permanent-popup'
    }).setContent(content);
}

function loadDataForViewport() {
    // When Top N filter is active, skip loading full viewport data.
    if(activeFilters.topN) {
        markersLayer.clearLayers();
        linesLayer.clearLayers();
        return;
    }
    var bounds = map.getBounds();
    var params = {
        min_lat: bounds.getSouth(),
        max_lat: bounds.getNorth(),
        min_lon: bounds.getWest(),
        max_lon: bounds.getEast(),
        offset: 0,
        limit: 500
    };
    
    // Special case: if "Unmatched" and OSM node type are both selected,
    // we need to include 'osm' in the stop_filter because OSM nodes are stored with stop_type='osm'
    var needsOsmType = activeFilters.stopType.includes('unmatched') && 
                       activeFilters.nodeType.includes('osm') && 
                       !activeFilters.stopType.includes('osm');
    
    if(activeFilters.stopType.length > 0) {
        var stopFilterTypes = [...activeFilters.stopType];
        if (needsOsmType) {
            stopFilterTypes.push('osm');
        }
        params.stop_filter = stopFilterTypes.join(',');
    }
    
    if(activeFilters.matchMethods.length > 0) {
        params.match_method = activeFilters.matchMethods.join(',');
    }
    
    // Add station filter values and type if available
    if(activeFilters.station.length > 0) {
        params.station_filter = activeFilters.station.join(',');
        params.filter_types = activeFilters.stationTypes.join(',');
        params.route_directions = activeFilters.routeDirections.join(',');
    }
    
    $.getJSON("/api/data", params, function(data) {
         markersLayer.clearLayers();
         // Only clear non-manual match lines
         linesLayer.eachLayer(function(layer) {
             if (!layer.options || !layer.options.isManualMatch) {
                 linesLayer.removeLayer(layer);
             }
         });
         stopsById = {};
         
         // Check if node type filtering is active
         var showAtlasNodes = activeFilters.nodeType.length === 0 || activeFilters.nodeType.indexOf("atlas") !== -1;
         var showOSMNodes = activeFilters.nodeType.length === 0 || activeFilters.nodeType.indexOf("osm") !== -1;

         // If "Unmatched" and OSM are both selected, we need special handling
         var showUnmatchedOSM = activeFilters.stopType.includes('unmatched') && showOSMNodes;
         
         // First process data to find reverse matches (OSM nodes with multiple ATLAS stops)
         var osmNodeToAtlasMatches = {};
         
         // Process all matched ATLAS stops to find OSM nodes with multiple ATLAS matches
         data.forEach(function(stop) {
             if (stop.stop_type === 'matched' && stop.osm_matches && Array.isArray(stop.osm_matches)) {
                 stop.osm_matches.forEach(function(osmMatch) {
                     if (!osmMatch.osm_node_id) return;
                     
                     // Initialize the array if it doesn't exist
                     if (!osmNodeToAtlasMatches[osmMatch.osm_node_id]) {
                         osmNodeToAtlasMatches[osmMatch.osm_node_id] = {
                             atlas_matches: [],
                             osm_data: osmMatch
                         };
                     }
                     
                     // Add this ATLAS stop to the OSM node's matches
                     osmNodeToAtlasMatches[osmMatch.osm_node_id].atlas_matches.push({
                         sloid: stop.sloid,
                         uic_ref: stop.uic_ref,
                         atlas_designation: stop.atlas_designation,
                         atlas_designation_official: stop.atlas_designation_official,
                         atlas_business_org_abbr: stop.atlas_business_org_abbr,
                         atlas_lat: stop.atlas_lat,
                         atlas_lon: stop.atlas_lon,
                         distance_m: osmMatch.distance_m,
                         match_type: osmMatch.match_type,
                         routes_atlas: stop.routes_atlas
                     });
                 });
             }
         });
         
         data.forEach(function(stop) {
            stopsById[stop.id] = stop;
            
            if(stop.stop_type === 'matched') {
                // Unified popup approach for matched stops
                let unifiedPopup = null;
                
                if(stop.osm_matches && Array.isArray(stop.osm_matches)) {
                    // Handle ATLAS stop with multiple OSM matches
                    if(showAtlasNodes && stop.atlas_lat && stop.atlas_lon) {
                        unifiedPopup = createPopupWithOptions(createUnifiedPopupContent(stop));
                        
                        var atlasMarker = L.circleMarker([parseFloat(stop.atlas_lat), parseFloat(stop.atlas_lon)], 
                            { color: "green", radius: 6 })
                            .bindPopup(unifiedPopup);
                        markersLayer.addLayer(atlasMarker);
                    }
                    
                    // Display all OSM matches and connecting lines
                    if(showOSMNodes) {
                        stop.osm_matches.forEach(function(osm_match) {
                            if(osm_match.osm_lat && osm_match.osm_lon) {
                                var isPublicTransportPlatform = osm_match.osm_public_transport === 'platform';
                                
                                // Check if this OSM node has multiple ATLAS matches
                                const nodeId = osm_match.osm_node_id;
                                const hasMultipleAtlasMatches = 
                                    osmNodeToAtlasMatches[nodeId] && 
                                    osmNodeToAtlasMatches[nodeId].atlas_matches.length > 1;
                                
                                let osmNodeMarker;
                                
                                // If this OSM node has multiple ATLAS matches, use a special popup for it
                                if (hasMultipleAtlasMatches) {
                                    // Create a special data structure for the OSM node with its ATLAS matches
                                    const osmWithMatches = {
                                        id: osm_match.osm_id,
                                        stop_type: 'matched',
                                        is_osm_node: true,
                                        osm_node_id: osm_match.osm_node_id,
                                        osm_name: osm_match.osm_name,
                                        osm_uic_name: osm_match.osm_uic_name,
                                        osm_local_ref: osm_match.osm_local_ref,
                                        osm_network: osm_match.osm_network,
                                        osm_operator: osm_match.osm_operator,
                                        osm_public_transport: osm_match.osm_public_transport,
                                        osm_lat: osm_match.osm_lat,
                                        osm_lon: osm_match.osm_lon,
                                        lat: osm_match.osm_lat,
                                        lon: osm_match.osm_lon,
                                        routes_osm: osm_match.routes_osm,
                                        atlas_matches: osmNodeToAtlasMatches[nodeId].atlas_matches
                                    };
                                    
                                    osmNodeMarker = createOsmMarker(
                                        parseFloat(osm_match.osm_lat), 
                                        parseFloat(osm_match.osm_lon), 
                                        isPublicTransportPlatform,
                                        'blue'
                                    );
                                    
                                    // Create a unified popup with the OSM-centric view
                                    osmNodeMarker.bindPopup(createPopupWithOptions(createUnifiedPopupContent(osmWithMatches)));
                                }
                                // Otherwise use the standard unified popup
                                else {
                                    // If we haven't created the unified popup yet (because ATLAS nodes are hidden)
                                    if (!unifiedPopup) {
                                        unifiedPopup = createPopupWithOptions(createUnifiedPopupContent(stop));
                                    }
                                    
                                    osmNodeMarker = createOsmMarker(
                                        parseFloat(osm_match.osm_lat), 
                                        parseFloat(osm_match.osm_lon), 
                                        isPublicTransportPlatform,
                                        'blue'
                                    );
                                    
                                    osmNodeMarker.bindPopup(unifiedPopup);
                                }
                                
                                markersLayer.addLayer(osmNodeMarker);
                                
                                // Store the OSM node in the global stopsById for later reference
                                stopsById[osm_match.osm_id] = {
                                    id: osm_match.osm_id,
                                    osm_node_id: osm_match.osm_node_id,
                                    osm_local_ref: osm_match.osm_local_ref,
                                    osm_name: osm_match.osm_name,
                                    osm_uic_name: osm_match.osm_uic_name,
                                    uic_ref: stop.uic_ref,  // Use parent stop's uic_ref
                                    osm_network: osm_match.osm_network,
                                    osm_operator: osm_match.osm_operator,
                                    osm_lat: osm_match.osm_lat,
                                    osm_lon: osm_match.osm_lon,
                                    lat: osm_match.osm_lat,
                                    lon: osm_match.osm_lon,
                                    distance_m: osm_match.distance_m,
                                    match_type: osm_match.match_type || stop.match_type,  // Use the match_type
                                    stop_type: 'matched',
                                    routes_osm: osm_match.routes_osm
                                };
                                
                                // Add connecting line if both ATLAS and OSM nodes are visible
                                if(showAtlasNodes && stop.atlas_lat && stop.atlas_lon) {
                                    var line = L.polyline([
                                        [parseFloat(stop.atlas_lat), parseFloat(stop.atlas_lon)],
                                        [parseFloat(osm_match.osm_lat), parseFloat(osm_match.osm_lon)]
                                    ], { color: "green" });
                                    linesLayer.addLayer(line);
                                }
                            }
                        });
                    }
                } 
                else if(stop.atlas_lat && stop.atlas_lon && stop.osm_lat && stop.osm_lon) {
                    // Handle regular matched stops (one ATLAS to one OSM)
                    unifiedPopup = createPopupWithOptions(createUnifiedPopupContent(stop));
                    
                    if(showAtlasNodes) {
                        var atlasMarker = L.circleMarker([parseFloat(stop.atlas_lat), parseFloat(stop.atlas_lon)], 
                            { color: "green", radius: 6 })
                            .bindPopup(unifiedPopup);
                        markersLayer.addLayer(atlasMarker);
                    }
                    
                    if(showOSMNodes) {
                        var isPublicTransportPlatform = stop.osm_public_transport === 'platform';
                        var osmMarker = createOsmMarker(
                            parseFloat(stop.osm_lat), 
                            parseFloat(stop.osm_lon), 
                            isPublicTransportPlatform,
                            'blue'
                        );
                        osmMarker.bindPopup(unifiedPopup);
                        markersLayer.addLayer(osmMarker);
                    }
                    
                    if(showAtlasNodes && showOSMNodes) {
                        var line = L.polyline([
                            [parseFloat(stop.atlas_lat), parseFloat(stop.atlas_lon)],
                            [parseFloat(stop.osm_lat), parseFloat(stop.osm_lon)]
                        ], { color: "green" });
                        linesLayer.addLayer(line);
                    }
                }
            } else if (stop.stop_type === 'station') {
                // Only show station-matched ATLAS stops if ATLAS nodes should be displayed
                if (showAtlasNodes) {
                var popupContent = "<strong>Station-Matched ATLAS Stop:</strong><br>" +
                                   "Sloid: " + stop.sloid + "<br>" +
                                   "Designation: " + (stop.atlas_designation || stop.atlas_designation_official || "N/A") + "<br>" +
                                   "Official Name: " + (stop.atlas_designation_official || "N/A") + "<br>" +
                                   "Business Org: " + (stop.atlas_business_org_abbr || "N/A") + "<br>" +
                                   "UIC Ref: " + createFilterLink(stop.uic_ref, 'station') + "<br>" +
                                   "Coordinates: (" + (stop.atlas_lat || stop.lat) + ", " + (stop.atlas_lon || stop.lon) + ")<br>" +
                                   createCollapsible("Routes", formatRoutesDisplay(stop.routes_atlas, [], false)) +
                                   "<button onclick='filterByStation(" + stop.id + ", \"atlas\")'>Filter by station</button>" +
                                   " <button onclick='manualMatchSelect(" + stop.id + ", \"atlas\")'>Match To</button>";
                var marker = L.circleMarker([parseFloat(stop.atlas_lat || stop.lat), parseFloat(stop.atlas_lon || stop.lon)], { color: "orange", radius: 6 })
                    .bindPopup(createPopupWithOptions(popupContent));
                markersLayer.addLayer(marker);
                }
            } else if (stop.stop_type === 'unmatched') {
                // Only show unmatched ATLAS stops if ATLAS nodes should be displayed
                if (showAtlasNodes) {
                    // Use atlas-match class and structure, ADD unmatched class
                    let popupHtml = `
                        <div class="atlas-match unmatched">
                            <h5>Unmatched ATLAS Stop</h5>
                            <table class="popup-table">
                                <tr><td>Sloid:</td><td>${stop.sloid}</td></tr>
                                <tr><td>Designation:</td><td>${stop.atlas_designation || stop.atlas_designation_official || 'N/A'}</td></tr>
                                <tr><td>Business Org:</td><td>${stop.atlas_business_org_abbr || 'N/A'}</td></tr>
                                <tr><td>UIC Ref:</td><td>${createFilterLink(stop.uic_ref, 'station')}</td></tr>
                            </table>
                            <div class="route-section">
                                ${createCollapsible("Routes", formatRoutesDisplay(stop.routes_atlas, [], false))}
                            </div>
                            <button onclick='manualMatchSelect(${stop.id}, "atlas")'>Match To</button>
                        </div>
                    `;
                    var marker = L.circleMarker([stop.lat, stop.lon], { color: "red", radius: 6 })
                        .bindPopup(createPopupWithOptions(popupHtml)); // Use the new html
                    markersLayer.addLayer(marker);
                }
            } else if (stop.stop_type === 'osm') {
                // Show unmatched OSM nodes using osm-match class and structure
                if (showOSMNodes && (!activeFilters.stopType.includes('unmatched') || showUnmatchedOSM)) {
                    var isPublicTransportPlatform = stop.osm_public_transport === 'platform';
                    // Build HTML directly using osm-match class, ADD unmatched class
                    let popupHtml = `
                        <div class="osm-match unmatched">
                            <h5>Unmatched OSM Node</h5>
                            <table class="popup-table">
                                <tr><td>Name:</td><td>${stop.osm_name || 'N/A'}</td></tr>
                                <tr><td>UIC Name:</td><td>${stop.osm_uic_name || 'N/A'}</td></tr>
                                <tr><td>UIC Ref:</td><td>${createFilterLink(stop.uic_ref, 'station')}</td></tr>
                                <tr><td>Network:</td><td>${stop.osm_network || 'N/A'}</td></tr>
                                <tr><td>Operator:</td><td>${stop.osm_operator || 'N/A'}</td></tr>
                                <tr><td>Type:</td><td>${stop.osm_public_transport || 'N/A'}</td></tr>
                                <tr><td>Node ID:</td><td>${stop.osm_node_id || 'N/A'}</td></tr>
                                <tr><td>Local Ref:</td><td>${stop.osm_local_ref || 'N/A'}</td></tr>
                            </table>
                            <div class="route-section">
                                ${createCollapsible("Routes", formatRoutesDisplay([], stop.routes_osm, true))}
                            </div>
                             <button onclick='filterByStation(${stop.id}, "osm")'>Filter by station</button>
                             <button onclick='manualMatchSelect(${stop.id}, "osm")'>Match To</button>
                        </div>
                    `;
                    var marker = createOsmMarker(
                        parseFloat(stop.lat),
                        parseFloat(stop.lon),
                        isPublicTransportPlatform,
                        'gray'
                    );
                    marker.bindPopup(createPopupWithOptions(popupHtml)); // Use the new html
                    markersLayer.addLayer(marker);
                }
            } else {
                // For any other types, we should respect the node type filters
                // This is a fallback case
                if (showAtlasNodes && showOSMNodes) {
                var popupContent = "Stop: " + stop.sloid;
                var marker = L.circleMarker([stop.lat, stop.lon], { color: "black", radius: 6 })
                    .bindPopup(createPopupWithOptions(popupContent));
                markersLayer.addLayer(marker);
                }
            }
         });
         
         // Now add all the lines for OSM nodes with multiple ATLAS matches
         if(showAtlasNodes && showOSMNodes) {
             Object.keys(osmNodeToAtlasMatches).forEach(function(osmNodeId) {
                 const osmData = osmNodeToAtlasMatches[osmNodeId];
                 
                 // Only process if the OSM node has multiple ATLAS matches
                 if (osmData.atlas_matches.length > 1 && osmData.osm_data) {
                     const osmLat = parseFloat(osmData.osm_data.osm_lat);
                     const osmLon = parseFloat(osmData.osm_data.osm_lon);
                     
                     // Add lines from this OSM node to all its ATLAS matches
                     osmData.atlas_matches.forEach(function(atlasMatch) {
                         if (atlasMatch.atlas_lat && atlasMatch.atlas_lon) {
                             var line = L.polyline([
                                 [parseFloat(atlasMatch.atlas_lat), parseFloat(atlasMatch.atlas_lon)],
                                 [osmLat, osmLon]
                             ], { color: "purple" });
                             linesLayer.addLayer(line);
                         }
                     });
                 }
             });
         }
         
         // Redraw manual match lines after loading data
         redrawManualMatchLines();
    });
}

// Helper function to create stop popup content - kept for backwards compatibility
function createStopPopupContent(stop, stopType) {
    let content = "";
    
    if (stopType === "matched_atlas") {
        content = "<strong>Matched ATLAS Stop:</strong><br>" +
                  "ATLAS Stop: " + stop.sloid + "<br>" +
                  "Designation: " + (stop.atlas_designation || "N/A") + "<br>" +
                  "Official Name: " + (stop.atlas_designation_official || "N/A") + "<br>" +
                  "Business Org: " + (stop.atlas_business_org_abbr || "N/A") + "<br>" +
                  "UIC Ref: " + createFilterLink(stop.uic_ref, 'station') + "<br>" +
                  "ATLAS Coords: (" + stop.atlas_lat + ", " + stop.atlas_lon + ")<br>" +
                  "OSM Coords: (" + stop.osm_lat + ", " + stop.osm_lon + ")<br>" +
                  "Distance: " + (stop.distance_m ? parseFloat(stop.distance_m).toFixed(1) + " m" : "N/A") + "<br>" +
                  "Match Type: " + (stop.match_type || "N/A") + "<br>" +
                  createCollapsible("Routes", formatRoutesDisplay(stop.routes_atlas, stop.routes_osm, false));
    } else if (stopType === "matched_atlas_multi") {
        content = "<strong>Matched ATLAS Stop:</strong><br>" +
                  "ATLAS Stop: " + stop.sloid + "<br>" +
                  "Designation: " + (stop.atlas_designation || "N/A") + "<br>" +
                  "Official Name: " + (stop.atlas_designation_official || "N/A") + "<br>" +
                  "Business Org: " + (stop.atlas_business_org_abbr || "N/A") + "<br>" +
                  "UIC Ref: " + createFilterLink(stop.uic_ref, 'station') + "<br>" +
                  "ATLAS Coords: (" + stop.atlas_lat + ", " + stop.atlas_lon + ")<br>" +
                  "Matched to " + stop.osm_matches.length + " OSM nodes<br>" +
                  createCollapsible("Routes", formatRoutesDisplay(stop.routes_atlas, stop.routes_osm, false));
    } else if (stopType === "matched_osm") {
        content = "<strong>OSM Node:</strong><br>" +
                  "Name: " + (stop.osm_name || "N/A") + "<br>" +
                  "UIC Name: " + (stop.osm_uic_name || "N/A") + "<br>" +
                  "UIC Ref: " + createFilterLink(stop.uic_ref, 'station') + "<br>" +
                  "Local Ref: " + (stop.osm_local_ref || "N/A") + "<br>" +
                  "Network: " + (stop.osm_network || "N/A") + "<br>" +
                  "Operator: " + (stop.osm_operator || "N/A") + "<br>" +
                  (stop.osm_public_transport || "N/A") + "<br>" +
                  "Node ID: " + (stop.osm_node_id || "N/A") + "<br>" +
                  "Distance: " + (stop.distance_m ? parseFloat(stop.distance_m).toFixed(1) + " m" : "N/A") + "<br>" +
                  "Match Type: " + (stop.match_type || "N/A") + "<br>" +
                  createCollapsible("Routes", formatRoutesDisplay(stop.routes_atlas, stop.routes_osm, true));
    } else if (stopType === "unmatched_atlas") {
        content = "<strong>Unmatched ATLAS Stop:</strong><br>" +
                  "Sloid: " + stop.sloid + "<br>" +
                  "Designation: " + (stop.atlas_designation || stop.atlas_designation_official || "N/A") + "<br>" +
                  "Business Org: " + (stop.atlas_business_org_abbr || "N/A") + "<br>" +
                  "UIC Ref: " + createFilterLink(stop.uic_ref, 'station') + "<br>" +
                  createCollapsible("Routes", formatRoutesDisplay(stop.routes_atlas, [], false)) +
                  " <button onclick='manualMatchSelect(" + stop.id + ", \"atlas\")'>Match To</button>";
    } else if (stopType === "unmatched_osm") {
        content = "<strong>Unmatched OSM Node:</strong><br>" +
                  "Name: " + (stop.osm_name || "N/A") + "<br>" +
                  "UIC Name: " + (stop.osm_uic_name || "N/A") + "<br>" +
                  "UIC Ref: " + createFilterLink(stop.uic_ref, 'station') + "<br>" +
                  "Network: " + (stop.osm_network || "N/A") + "<br>" +
                  "Operator: " + (stop.osm_operator || "N/A") + "<br>" +
                  (stop.osm_public_transport || "N/A") + "<br>" +
                  "Node ID: " + (stop.osm_node_id || "N/A") + "<br>" +
                  "Local Ref: " + (stop.osm_local_ref || "N/A") + "<br>" +
                  createCollapsible("Routes", formatRoutesDisplay([], stop.routes_osm, true)) +
                  " <button onclick='manualMatchSelect(" + stop.id + ", \"osm\")'>Match To</button>";
    }
    
    return content;
}

$(document).ready(function(){
    initMap();
    loadDataForViewport();
    
    // Attach change handlers to filter checkboxes and input elements.
    $('.filter-stop-type, .filter-node-type, .filter-match-method, .filter-distance-method, .filter-route-method').on('change', updateActiveFilters);
    
    // Setup Add button for station filter
    $('#addStationFilter').on('click', function(e) {
        e.preventDefault();
        addTextFilter();
    });
    
    // Setup Add button for route filter
    $('#addRouteFilter').on('click', function(e) {
        e.preventDefault();
        addRouteFilter();
    });
    
    // Also allow adding by pressing Enter in the text inputs
    $('#stationFilter').on('keypress', function(e) {
        if (e.which === 13) { // Enter key
            e.preventDefault();
            addTextFilter();
        }
    });
    
    $('#routeIdFilter, #directionFilter').on('keypress', function(e) {
        if (e.which === 13) { // Enter key
            e.preventDefault();
            addRouteFilter();
        }
    });
    
    // Add event listener for filter type selector to toggle input fields
    $('#filterTypeSelector').on('change', function() {
        toggleFilterInputs();
    });
    
    // Initialize display based on default filter type
    toggleFilterInputs();
    
    // Toggle button for Top N filter
    $('#toggleTopNBtn').on('click', function(){
        if(activeFilters.topN) {
            activeFilters.topN = null;
            $(this).text("Activate Top N Filter");
            topNLayer.clearLayers();
            $('#topNDistancesMessage').empty();
            loadDataForViewport();
            updateFiltersUI();
        } else {
            var n = parseInt($('#topDistance').val());
            if(n > 0) {
                activeFilters.topN = n;
                $(this).text("Remove Top N Filter");
                loadTopNMatches();
                updateFiltersUI();
            } else {
                alert("Please enter a valid number for Top N filter.");
            }
        }
    });
    
    // Initialize report generation and manual matching features
    initReportGeneration();
    initManualMatching();
});
